/*
Copyright (c) 2000-2002, Jelle Kok, University of Amsterdam
All rights reserved.

Redistribution and use in source and binary forms, with or without 
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
list of conditions and the following disclaimer. 

2. Redistributions in binary form must reproduce the above copyright notice, 
this list of conditions and the following disclaimer in the documentation 
and/or other materials provided with the distribution. 

3. Neither the name of the University of Amsterdam nor the names of its 
contributors may be used to endorse or promote products derived from this 
software without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE 
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
/*! \file WorldModelHighLevel.C
<pre>
<b>File:</b>          WorldModelHighLevel.C
<b>Project:</b>       Robocup Soccer Simulation Team: UvA Trilearn
<b>Authors:</b>       Jelle Kok
<b>Created:</b>       12/02/2001
<b>Last Revision:</b> $ID$
<b>Contents:</b>      class definitions of WorldModel. This class contains
               methods that reason about the world model on a higher
               level and return more abstract information about the current
               world state.
<hr size=2>
<h2><b>Changes</b></h2>
<b>Date</b>             <b>Author</b>          <b>Comment</b>
12/02/2001       Jelle Kok       Initial version created
</pre>
*/

#include<list>            // needed for list<double>
#include<stdio.h>         // needed for printf
#include "WorldModel.h"


/*! This method returns the number of visible objects that are part of the
    object set 'set' and located in the rectangle 'rect'. When no
    rectangle is defined (rect=NULL) the whole field is taken into account. Only
    objects with a confidence value higher than the threshold defined in
    PlayerSettings are taken into consideration.
    \param set ObjectSetT from which objects are taken into consideration
    \param rect Rectangle in which objects are counted (default NULL)
    \return number of objects in Rectangle 'rect'. */
int WorldModel::getNrInSetInRectangle( ObjectSetT set,
                                       Rectangle *rect)
{
  double dConfThr = PS->getPlayerConfThr();
  int    iNr      = 0;
  int    iIndex;

  for( ObjectT o = iterateObjectStart( iIndex, set, dConfThr );
       o != OBJECT_ILLEGAL;
       o = iterateObjectNext ( iIndex, set, dConfThr ) )
  {
    if( rect == NULL || rect->isInside( getGlobalPosition( o ) ) )
      iNr++;
  }
  iterateObjectDone( iIndex );
  return iNr;
}

/*! This method returns the number of objects that are within the circle 'c'
    Only objects are taken into account that are part of the set 'set' and
    have a confidence higher than the threshold defined in PlayerSettings.
    \param c circle in which objects should be located to be counted
    \return number of objects from 'set' in circle 'c'*/
int WorldModel::getNrInSetInCircle( ObjectSetT set, Circle c )
{
  double dConfThr = PS->getPlayerConfThr();
  int    iNr      = 0;
  int    iIndex;

  for( ObjectT o = iterateObjectStart( iIndex, set, dConfThr );
       o != OBJECT_ILLEGAL;
       o = iterateObjectNext ( iIndex, set, dConfThr ) )
  {
    if( c.isInside( getGlobalPosition( o ) ) )
      iNr++;
  }
  iterateObjectDone( iIndex );

  return iNr;
}

/*! This method returns the number of objects in the specified cone.
    A cone is like a piece of a pie, in which 'start' is
    the center of the pie, 'end' is the edge of the pie and 'dWidth' is the
    half width of the piece after distance 1. Only objects are taken into
    consideration that are within the set 'set' and have a confidence higher
    than the threshold defined in PlayerSettings.
    \param set ObjectSetT of which objects are taken into consideration
    \param dWidth half width of the cone after distance 1.0
    \param start center of the cone
    \param end position that is the end of the cone.
    \return number of objects part of 'set' and located in this cone. */
int WorldModel::getNrInSetInCone( ObjectSetT set, double dWidth,
                                      VecPosition start , VecPosition end )
{
  double      dConfThr   = PS->getPlayerConfThr();
  int         iNr        = 0;
  int         iIndex;
  Line        line       = Line::makeLineFromTwoPoints( start, end );
  VecPosition posOnLine;
  VecPosition posObj;

  for( ObjectT o = iterateObjectStart( iIndex, set, dConfThr );
       o != OBJECT_ILLEGAL;
       o = iterateObjectNext ( iIndex, set, dConfThr ) )
  {
    //posObj    = getGlobalPosition( o );
    posObj    = predictPlayerPos(3, o );
    posOnLine = line.getPointOnLineClosestTo( posObj );
    // whether posOnLine lies in cone is checked by three constraints
    // - does it lie in triangle (to infinity)
    // - lies between start and end (and thus not behind me)
    // - does it lie in circle
    if(posOnLine.getDistanceTo(posObj) < dWidth*posOnLine.getDistanceTo(start)
       && line.isInBetween( posOnLine, start, end )
       && start.getDistanceTo( posObj ) < start.getDistanceTo( end ) )
        iNr++;
  }
  iterateObjectDone( iIndex );
  return iNr;
}

/*! This method returns the object type of the closest object to the ObjectT
    that is supplied as the second argument. Only objects are taken into account
    that are part of the set 'set' and have a confidence higher than the
    supplied threshold. If no threshold is supplied, the threshold defined in
    PlayerSettings is used.
    \param set ObjectSetT which denotes objects taken into consideration
    \param objTarget ObjectT that represent the type of the object to compare to
    \param dDist will be filled with the closest distance
    \param dConfThr minimum confidence threshold for the objects in 'set'
    \return ObjectType that is closest to o */
ObjectT WorldModel::getClosestInSetTo( ObjectSetT set, ObjectT objTarget,
                                        double *dDist, double dConfThr )
{
  if( dConfThr == -1.0 ) dConfThr      = PS->getPlayerConfThr();
  ObjectT     closestObject = OBJECT_ILLEGAL;
  double      dMinMag       = 1000.0;
  VecPosition v;
  int         iIndex;

  for( ObjectT o = iterateObjectStart( iIndex, set, dConfThr );
       o != OBJECT_ILLEGAL;
       o = iterateObjectNext ( iIndex, set, dConfThr ) )
  {
    if( o != objTarget )               // do not include target object
    {
      v = getGlobalPosition( objTarget ) - getGlobalPosition( o );
      if( v.getMagnitude() < dMinMag )
      {
        dMinMag       = v.getMagnitude();
        closestObject = o;
      }
    }
  }

  iterateObjectDone( iIndex );
  if( dDist != NULL )
    *dDist = dMinMag;
  return closestObject;
}

/*! This method returns the ojbect type of the closest object to the specified
    position and that is part of the object set 'set' with a confidence
    higher than the
    supplied threshold. If no threshold is supplied, the threshold defined in
    PlayerSettings is used.
    \param set ObjectSetT which denotes objects taken into consideration
    \param pos position to which player should be compared
    \param dDist will be filled with the distance between pos and closest object
    \param dConfThr minimum confidence threshold for the objects in 'set'
    \return ObjectT representing object that is closest to pos */
ObjectT WorldModel::getClosestInSetTo( ObjectSetT set, VecPosition pos,
                                       double *dDist,  double dConfThr )
{
  ObjectT     closestObject   = OBJECT_ILLEGAL;
  double      dMinMag         = 1000.0;
  VecPosition v;
  int         iIndex;

  if( dConfThr == -1.0 ) dConfThr      = PS->getPlayerConfThr();
  for( ObjectT o = iterateObjectStart( iIndex, set, dConfThr );
       o != OBJECT_ILLEGAL;
       o = iterateObjectNext ( iIndex, set, dConfThr ) )
  {
    v = pos - getGlobalPosition( o );
    if( v.getMagnitude() < dMinMag )
    {
      dMinMag         = v.getMagnitude();
      closestObject   = o;
    }
  }
  iterateObjectDone( iIndex );
  if( dDist != NULL )
    *dDist = dMinMag;
  return closestObject;
}

/*! This method returns the closest object in 'set' to the line 'l'. The
    projection p of the global position of this object on the line 'l' should
    lie between pos1 and pos2. After the method is finished, it returns this
    object and the last two arguments of this method are set to the the distance
    between the object and p and the distance from pos1 to p respectively.
    \param set ObjectSetT which denotes objects taken into consideration
    \param l line to which opponents should be projected
    \param pos1 minimum allowed projection point
    \param pos2 maximum allowed projection point
    \param dDistObjToLine will contain distance from opponent to line l
    \param dDistPos1ToPoint will contain distance from pos1 to projection point
    opponent on line l
    \return object type of closest object to line l */
ObjectT WorldModel::getClosestInSetTo( ObjectSetT set, Line l,
                              VecPosition pos1, VecPosition pos2,
                              double *dDistObjToLine, double *dDistPos1ToPoint)
{
  VecPosition posObj;
  VecPosition posOnLine;
  double      dConfThr  = PS->getPlayerConfThr();
  ObjectT     obj       = OBJECT_ILLEGAL;
  double      dDist     = 1000.0;
  double      dMinDist  = 1000.0;
  double      dDistPos1 = 1000.0;
  int         iIndex;

  for( ObjectT o = iterateObjectStart( iIndex, set, dConfThr );
       o != OBJECT_ILLEGAL;
       o = iterateObjectNext ( iIndex, set, dConfThr ) )
  {
    posObj    = getGlobalPosition( o );
    posOnLine = l.getPointOnLineClosestTo( posObj );
    dDist     = posObj.getDistanceTo( posOnLine );
    if( l.isInBetween( posOnLine, pos1, pos2 ) && dDist < dMinDist )
    {
      dMinDist  = dDist;
      obj       = o;
      dDistPos1 = pos1.getDistanceTo( posOnLine );
    }
  }
  iterateObjectDone( iIndex );
  if( dDistObjToLine != NULL )
    *dDistObjToLine = dMinDist;
  if( dDistPos1ToPoint != NULL )
    *dDistPos1ToPoint = dDistPos1;

  return obj;
}

/*! This method returns the object type of the closest object relative to
    the agent. Only objects are taken into account that are part of the set
    'set'.
    \param set ObjectSetT which denotes objects taken into consideration
    \param dDist will be filled with the closest relative distance
    \return ObjectType that is closest to the agent*/
ObjectT WorldModel::getClosestRelativeInSet( ObjectSetT set, double *dDist )
{
  ObjectT     closestObject = OBJECT_ILLEGAL;
  double      dMinMag       = 1000.0;
  int         iIndex;

  for( ObjectT o = iterateObjectStart( iIndex, set, 1.0 );
       o != OBJECT_ILLEGAL;
       o = iterateObjectNext ( iIndex, set, 1.0 ) )
  {
    if( getRelativeDistance( o ) < dMinMag )
    {
      dMinMag       = getRelativeDistance( o );
      closestObject = o;
    }
  }

  iterateObjectDone( iIndex );
  if( dDist != NULL )
    *dDist = dMinMag;
  return closestObject;
}

/*! This method returns the object type of the second closest object to the
    object type that is supplied as the second argument. Only objects are taken
    into account within set 'set' and with a confidence higher than the
    supplied threshold. If no threshold is supplied, the threshold defined in
    PlayerSettings is used.
    \param set ObjectSetT which denotes objects taken into consideration
    \param obj ObjectT that represent the type of the object to check
    \param dDist will be filled with the distance to this player.
    \param dConfThr minimum confidence threshold for the objects in 'set'
    \return ObjectType that is second closest to obj */
ObjectT WorldModel::getSecondClosestInSetTo ( ObjectSetT set, ObjectT obj,
                                              double *dDist,  double dConfThr )
{
  VecPosition v;
  ObjectT     closestObject       = OBJECT_ILLEGAL;
  ObjectT     secondClosestObject = OBJECT_ILLEGAL;
  double      dMinMag             = 1000.0;
  double      dSecondMinMag       = 1000.0;
  int         iIndex;

  if( dConfThr == -1.0 ) dConfThr = PS->getPlayerConfThr();

  for( ObjectT o = iterateObjectStart( iIndex, set, dConfThr );
       o != OBJECT_ILLEGAL;
       o = iterateObjectNext ( iIndex, set, dConfThr ) )
  {
    if( o != obj )
    {
      v = getGlobalPosition( obj ) - getGlobalPosition( o );
      if( v.getMagnitude() < dMinMag )                  // closer then first
      {
        dSecondMinMag         = dMinMag;                // put first to second
        secondClosestObject   = closestObject;
        dMinMag               = v.getMagnitude();       // and this to first
        closestObject         = o;
      }
      else if( v.getMagnitude() < dSecondMinMag )       // between first and 2nd
      {
        dSecondMinMag         = v.getMagnitude();       // put this to second
        secondClosestObject   = o;
      }
    }
  }
  iterateObjectDone( iIndex );
  if( dDist != NULL )
    *dDist = dSecondMinMag;
  return secondClosestObject;
}

/*! This method returns the object type of the second closest object relative
    to the agent. Only objects are taken into account within set 'set' and
    which where seen in the last see message.
    \param set ObjectSetT which denotes objects taken into consideration
    \param dDist will be filled with the distance to this this object
    \return ObjectType that is second closest to the agent */
ObjectT WorldModel::getSecondClosestRelativeInSet( ObjectSetT set, double *dDist )
{
  ObjectT     closestObject       = OBJECT_ILLEGAL;
  ObjectT     secondClosestObject = OBJECT_ILLEGAL;
  double      dMinMag             = 1000.0;
  double      dSecondMinMag       = 1000.0;
  double      d;
  int         iIndex;

  for( ObjectT o = iterateObjectStart( iIndex, set, 1.0 );
       o != OBJECT_ILLEGAL;
       o = iterateObjectNext ( iIndex, set, 1.0 ) )
  {
    d = getRelativeDistance( o );
    if( d < dMinMag )                                 // closer then first
    {
      dSecondMinMag         = dMinMag;                // put first to second
      secondClosestObject   = closestObject;
      dMinMag               = d;                      // and this to first
      closestObject         = o;
    }
    else if( d < dSecondMinMag )                      // between first and 2nd
    {
      dSecondMinMag         = d;                      // put this to second
      secondClosestObject   = o;
    }
  }
  iterateObjectDone( iIndex );
  if( dDist != NULL )
    *dDist = dSecondMinMag;
  return secondClosestObject;
}


/*! This method returns the object type of the furthest object to the ObjectT
    that is supplied as the second argument. Only objects are taken into account
    that are part of the set 'set' and have a confidence higher than the
    supplied threshold. If no threshold is supplied, the threshold defined in
    PlayerSettings is used.
    \param set ObjectSetT which denotes objects taken into consideration
    \param o ObjectT that represent the type of the object to compare to
    \param dDist will be filled with the furthest distance
    \param dConfThr minimum confidence threshold for the objects in 'set'
    \return ObjectType that is furthest to o */
ObjectT WorldModel::getFurthestInSetTo( ObjectSetT set, ObjectT objTarget,
                                        double *dDist, double dConfThr )
{
  if( dConfThr == -1.0 ) dConfThr      = PS->getPlayerConfThr();

  ObjectT     furthestObject = OBJECT_ILLEGAL;
  double      dMaxMag       = -1000.0;
  VecPosition v;
  int         iIndex;

  for( ObjectT o = iterateObjectStart( iIndex, set, dConfThr );
       o != OBJECT_ILLEGAL;
       o = iterateObjectNext ( iIndex, set, dConfThr ) )
  {
    if( o != objTarget )
    {
      v = getGlobalPosition( objTarget ) - getGlobalPosition( o );
      if( v.getMagnitude() > dMaxMag )
      {
        dMaxMag        = v.getMagnitude();
        furthestObject = o;
      }
    }
  }
  iterateObjectDone( iIndex );
  if( dDist != NULL )
    *dDist = dMaxMag;
  return furthestObject;
}

/*! This method returns the type of the object that is located furthest
    relative to the agent. Only objects are taken into account
    that are part of the set 'set'.
    \param set ObjectSetT which denotes objects taken into consideration
    \param dDist will be filled with the furthest relative distance
    \return ObjectType that is furthest to the agent */
ObjectT WorldModel::getFurthestRelativeInSet( ObjectSetT set, double *dDist  )
{
  ObjectT     furthestObject = OBJECT_ILLEGAL;
  double      dMaxMag       = -1000.0;
  int         iIndex;

  for( ObjectT o = iterateObjectStart( iIndex, set, 1.0 );
       o != OBJECT_ILLEGAL;
       o = iterateObjectNext ( iIndex, set, 1.0 ) )
  {
    if( getRelativeDistance( o ) > dMaxMag )
    {
      dMaxMag        = getRelativeDistance( o );
      furthestObject = o;
    }
  }
  iterateObjectDone( iIndex );
  if( dDist != NULL )
    *dDist = dMaxMag;
  return furthestObject;
}

/*! This method returns the maximum distance object 'o' could have traveled
    since it was last seen. 
    \param o object type to check
    \return distance which object 'o' could have maximal traveled. */
double WorldModel::getMaxTraveledDistance( ObjectT o )
{
  return (getCurrentTime() - getTimeLastSeen( o ) )*SS->getPlayerSpeedMax();

}

/*! This method returns the fastest object to a specified object and fills the
    last argument with the predicted amount of cycles needed to intercept
    this object. Only objects within the set 'set' are taken into
    consideration and the objects have to have a confidence higher than the
    player confidence threshold defined in PlayerSettings.
    \param set ObjectSetT which denotes objects taken into consideration
    \param obj object type of object that should be intercepted
    \param iCyclesToIntercept will be filled with the amount of cycles needed
    \returns object that can intercept object obj fastest */
ObjectT WorldModel::getFastestInSetTo( ObjectSetT set, ObjectT obj,
                                        int *iCyclesToIntercept )
{
  double      dConfThr      = PS->getPlayerConfThr();
  ObjectT     fastestObject = OBJECT_ILLEGAL;
  int         iCycles       = -1;
  int         iCyclesToObj  = 100;
  int         iMinCycles    = 100;
  int         iIndex;
  VecPosition posObj;

  while( iCycles < iMinCycles && iCycles < 100)
  {
    iCycles    = iCycles + 1  ;
    iMinCycles = 100;
    posObj     = predictPosAfterNrCycles( obj, iCycles );

    for( ObjectT o = iterateObjectStart( iIndex, set, dConfThr );
       o != OBJECT_ILLEGAL;
       o = iterateObjectNext ( iIndex, set, dConfThr ) )
    {
     // if( getGlobalPosition(o).getDistanceTo(posObj)/SS->getPlayerSpeedMax()
     //         < iMinCycles )
     // {
     //     iCyclesToObj = (int)rint(getGlobalPosition(o).getDistanceTo(posObj)
     //                        /SS->getPlayerSpeedMax());
        iCyclesToObj = predictNrCyclesToPoint( o, posObj,
                                               PS->getPlayerWhenToTurnAngle() );
        if( iCyclesToObj < iMinCycles )
        {
          iMinCycles = iCyclesToObj;
          fastestObject = o;
        }
     // }
    }
    iterateObjectDone( iIndex );
  }

  if( iCyclesToIntercept != NULL )
    *iCyclesToIntercept  = iMinCycles;
  return fastestObject;
}

/*! This method maps the fastest players to a specified object and fills some
    arguments with the predicted amount of cycles needed to intercept
    this object. Only objects are taken into
    consideration those to have a confidence higher than the
    player confidence threshold defined in PlayerSettings.
    \param set ObjectSetT which denotes objects taken into consideration
    \param obj object type of object that should be intercepted
    \param iCyclesToIntercept will be filled with the amount of cycles needed
    \returns object that can intercept object obj fastest */
ObjectT WorldModel::getSecondFastestInSetTo( ObjectSetT set, ObjectT obj,
                                        int *iCyclesToIntercept )
{
  double      dConfThr      = PS->getPlayerConfThr();
  
  ObjectT     fastestObject = OBJECT_ILLEGAL;
  ObjectT     secondFastestObject = OBJECT_ILLEGAL;
  
  int         iCycles        = -1;
  
  int         iCyclesToObj  = 100;
  int         iMinCycles    = 100;
  int         iMinCycles2   = 100;
  int         iIndex;
  VecPosition posObj;

  
  // repeat until the number cycles to reach is equal to the number cycles to predict
  while( iCycles < iMinCycles2 && iCycles < 100)
  {
    iCycles    = iCycles + 1  ;
    iMinCycles    = 100;
    iMinCycles2   = 100;
    posObj     = predictPosAfterNrCycles( obj, iCycles );

    // find the second fastest  
    for( ObjectT o = iterateObjectStart( iIndex, set, dConfThr );
       o != OBJECT_ILLEGAL;
       o = iterateObjectNext ( iIndex, set, dConfThr ) )
    {
        // exclude the goalie
        if( SoccerTypes::isGoalie(o) )
        {
          continue;
        }
        iCyclesToObj = predictNrCyclesToPoint( o, posObj,
                                               PS->getPlayerWhenToTurnAngle() );
        if( iCyclesToObj < iMinCycles )
        {
          iMinCycles2 = iMinCycles;
          secondFastestObject = fastestObject;
          
          iMinCycles = iCyclesToObj;
          fastestObject = o;
          
        }
        else if( iCyclesToObj < iMinCycles2 )
        {
          iMinCycles2 = iCyclesToObj;
          secondFastestObject = o;

        }
    }
    iterateObjectDone( iIndex );
  }

  if( iCyclesToIntercept != NULL )
    *iCyclesToIntercept  = iMinCycles2;
  return secondFastestObject;
}

/*! This method returns the fastest object to another object that is currently
    located at position 'pos' and has velocity 'vel' that decays with a value
    'dDecay'. The last argument will be filled with the predicted amount of
    cycles needed to reach this object.
    \param set ObjectSetT which denotes objects taken into consideration
    \param pos current position of the object
    \param vel current velocity of the object
    \param dDecay decay value of the velocity of the object
    \param iCyclesToIntercept will be filled with the amount of cycles needed
    \returns object that can reach it fastest */
ObjectT WorldModel::getFastestInSetTo( ObjectSetT set, VecPosition pos,
                      VecPosition vel, double dDecay, int *iCyclesToIntercept)
{
  double  dConfThr      = PS->getPlayerConfThr();
  ObjectT fastestObject = OBJECT_ILLEGAL;
  int     iCycles       = 0;
  int     iCyclesToObj  = 100;
  int     iMinCycles    = 100;
  int     iIndex;

  while( iCycles < iMinCycles && iCycles < 100)
  {
    iCycles    = iCycles + 1  ;
    iMinCycles = 100;

    for( ObjectT o = iterateObjectStart( iIndex, set, dConfThr );
       o != OBJECT_ILLEGAL;
       o = iterateObjectNext ( iIndex, set, dConfThr ) )
    {
       if( getGlobalPosition(o).getDistanceTo(pos)/SS->getPlayerSpeedMax()
              < iMinCycles )
      {
        iCyclesToObj = (int)rint(getGlobalPosition(o).getDistanceTo(pos)
                             /SS->getPlayerSpeedMax() );
//        iCyclesToObj = predictNrCyclesToPoint( o, pos,
//                                               PS->getPlayerWhenToTurnAngle() );
//        if( iCyclesToObj < iMinCycles )
        {
          iMinCycles    = iCyclesToObj;
          fastestObject = o;
        }
      }
    }
    iterateObjectDone( iIndex );
    pos += vel;
    vel *= dDecay;
  }

  if( iCyclesToIntercept != NULL )
    *iCyclesToIntercept = iMinCycles;
  return fastestObject;
}

/*! This method returns the first empty spot in the set 'set'. The first empty
    spot is returned as the object which has a lower confidence than the
    threshold player_conf_thr defined in the PlayerSettings. This can be used
    when information of an unknown object is perceived. It is set on the first
    position where there is currently no information stored. If 'iUnknownPlayer'
    is specified, the range that corresponds to this unknown player is used
    to dermine the position.
    \param set ObjectSetT consisting of the objects to check
    \param iUnknownPlayer indicates the unknownplayer that has to be mapped
    \return object type of which currently no up to date information is stored*/
ObjectT WorldModel::getFirstEmptySpotInSet( ObjectSetT set, int iUnknownPlayer)
{
  int     iIndex;

  for( ObjectT o = iterateObjectStart( iIndex, set,  0.0 );
       o != OBJECT_ILLEGAL;
       o = iterateObjectNext ( iIndex, set, 0.0 ) )
  {
    if( getConfidence( o ) < PS->getPlayerConfThr() &&
        o != getAgentObjectType() && 
        ( iUnknownPlayer == -1 || UnknownPlayers[iUnknownPlayer].isInRange(o)))
      return o;
  }
  return OBJECT_ILLEGAL;
}


/*! This method returns a truth value that represents whether the object
    supplied as the first argument was seen in the last see message. If
    "touch" information was received, i.e. the object was very close to the
    agent but not in its view cone, false is also returned.
    \param ObjectT that represent the type of the object to check
    \return bool indicating whether o was seen in the last see message */
bool WorldModel::isVisible( ObjectT o )
{
  Object *object = getObjectPtrFromType( o );

  if( object != NULL &&
      object->getTimeLastSeen() == getTimeLastSeeMessage() )
    return true;

  return false;
}

/*! This method determines whether the ball is kicakble, i.e. the ball is in the
    kickable range of the agent (see ServerSettings). This value can be
    different for the different heterogeneous player types.
    \return bool indicating whether ball can be kicked. */
bool WorldModel::isBallKickable()
{
  return getRelativeDistance( OBJECT_BALL ) < SS->getMaximalKickDist();
}

/*! This method determines whether the ball is catchable. This only applies
    to a goalie. Three things are tested:
    - the ball hasn't been catched in catch_ban_cycles (see ServerSettings)
      before the current cycle
    - the relative distance to the ball is smaller than the length of the
      catchable area length of the goalie (see also ServerSettings)
    - the ball is in the (own) penalty area.

    \return true when above three constraints are met, false otherwise. */
bool WorldModel::isBallCatchable()
{
  return getTimeSinceLastCatch()            > SS->getCatchBanCycle()  &&
         getRelativeDistance( OBJECT_BALL ) < SS->getCatchableAreaL() &&
         isInOwnPenaltyArea( getBallPos() );
}

/*! This method checks whether the ball is currently heading towards our own
    goal. For the ball to be heading to our goal a few constraints must be met:
    - ball must be located in our penalty area  - !! some balls in the goal from outside penalty area!!!
    - line of ball heading must intersect with goal line within goal width +
    small constant.
    - ball must pass goal line within 20 cycles.

    If all these constraints are met true is returned, false otherwise
    \return bool indicating whether the ball is heading towards our own goal */
bool WorldModel::isBallHeadingToGoal(  )
{
  if( !isConfidenceGood( OBJECT_BALL ) ||
      //getBallPos().getX() >  - PENALTY_X + 5.0 )
      getBallPos().getX() >  - PENALTY_X + 10.0 )
    return false;

  // make line from ball heading and goal line
  Line l = Line::makeLineFromPositionAndAngle(getBallPos(), getBallDirection());
  Line l2= Line::makeLineFromTwoPoints( getPosOwnGoal(), getPosOwnGoal() +
                                                         VecPosition( 0, 10 ));

  // if intersection is outside goalwidth, not heading to goal
  VecPosition posIntersect = l.getIntersection( l2 );
  if( fabs(posIntersect.getY()) > SS->getGoalWidth()/2.0 + 3.0)
    return false;

  // check whether ball will be behind goal line within 20( 25 better) cycles.
  VecPosition pos    = getBallPos();
  int         iCycle = 1;
  while( pos.getX() > - PITCH_LENGTH/2.0 && iCycle < 25)
  {
    pos = predictPosAfterNrCycles( OBJECT_BALL, iCycle );
    iCycle ++;
  }

  return ( iCycle == 25 ) ? false : true;
}

/*! This method checks whether an opponent is currently making a cross pass.
It means that the ball will cross a line made from goal stick to 0 direction

*/

bool WorldModel::isCrossPass()
{
   //Line l2;
   VecPosition posGoalie = getAgentGlobalPosition();
   VecPosition posBall  = getBallPos();
   //if the ball is outside of defense area or too close to the central line
   if( !isConfidenceGood( OBJECT_BALL ) || posBall.getX() > -PITCH_LENGTH*0.5 +16.0 ||
      //getBallPos().getX() >  - PENALTY_X + 5.0 )
      fabs(posGoalie.getY()) < 5.0 || fabs(getBallPos().getY()) <  fabs(posGoalie.getY()))
    return false;


  // make line from ball heading and goal line
  Line l = Line::makeLineFromPositionAndAngle(getBallPos(), getBallDirection());
  Line l2= Line::makeLineFromPositionAndAngle( posGoalie, 0.0);


  // if intersection is outside 10 meters, not cross pass
  VecPosition posIntersect = l.getIntersection( l2 );
  if( posIntersect.getX() < -PITCH_LENGTH*0.5 || posIntersect.getX() > -PITCH_LENGTH*0.5 +10.0)
    return false;

  // check whether ball will be behind the line within 25 cycles.

  int         iCycle = 1;
  while( fabs(posBall.getY()) > fabs(posGoalie.getY()) && iCycle < 25)
  {
    posBall = predictPosAfterNrCycles( OBJECT_BALL, iCycle );
    iCycle ++;
  }

  return ( iCycle == 25 ) ? false : true;
}
/*! This method returns whether the ball is in our possesion. This is defined
     by the fact if the fastest player to the ball is a teammate or not.
     \return bool indicating whether a teammate is the fastest player to the
             ball. */

/*! The new implementation of this method gives three option - our possesion (1),
 opponent's possesion (-1) and not defined (0). The concept of the developers
 team. Implementation  S.Razykov*/
 
PossesionT WorldModel::isBallInOurPossesion( )
{
  int     iCyc1;
  int     iCyc2;

  // At first check dead ball situations
  if (isDeadBallUs() || isKickOffUs() )
    return POSSESION_US;                          // our possesion
  else if (isDeadBallThem() || isKickOffThem() )
    return POSSESION_THEM;                         // opponent's possesion
  if ( isBallKickable() )   // the player posseses the ball
    return POSSESION_US;
  // find a fastest teammate and a fastest opponent  
  ObjectT o1 = getFastestInSetTo( OBJECT_SET_TEAMMATES, OBJECT_BALL, &iCyc1 );
  ObjectT o2 = getFastestInSetTo( OBJECT_SET_OPPONENTS, OBJECT_BALL, &iCyc2 );
  
  //VecPosition posBall = predictPosAfterNrCycles(OBJECT_BALL, 2);

  if( o1 == OBJECT_ILLEGAL ) // it is impossible to find any teammate
  {
    //cout<<"No teammates"<<endl;
    return POSSESION_THEM;
  }
  else if(  o2 ==  OBJECT_ILLEGAL )  // it is impossible to find any opponent
  {
     //cout<<"No opponents"<<endl;
    return POSSESION_US;
  }
  //if( SoccerTypes::isTeammate( o ) )
    //return true;
    
  if( iCyc1 >iCyc2+2  ) // totally magic number. The teammate slower to the ball for 3 cycles or more
  {
    //cout<<"We Slower"<<endl;
    return POSSESION_THEM;
  }
  else if ( iCyc1+2<iCyc2 ) // totally magic number. The opponent slower to the ball for 3 cycles or more
  {
    //cout<<"We faster"<<endl;
    return POSSESION_US;
  }
  //else              // both the teammate and the opponent very close
  //{
    // if the ball is somewhere in the middle
    //if (getBallPos().isBetweenX(-PITCH_LENGTH/4, 3*PITCH_LENGTH/4))
    //{
    //  if (Ball.getGlobalVelocity().getX() < 0.0 ||!Ball.getPossesion() ) // the ball goes to left direction
    //   return false;
    //  else if (Ball.getGlobalVelocity().getX()>=  0.0 ||Ball.getPossesion() )
    //    return true;
    //  else if ( Ball.getPossesion() )
    //    return true;
    //  else
    //    return false;
    //}
    // closer to own goal
    //else //if( getBallPos().getX()<= -PITCH_LENGTH/4 ||
    // if (Ball.getPossesion() ) // we posessed the ball last cycle
    //    return true;
    // else
    //    return false;
  //}
  
   return POSSESION_NOTHING; // the ball possesion is not defined
}



/*! This method returns whether the ball lies in the own penalty area.
    \return bool indicating whether ball lies in own penalty area.     */
bool WorldModel::isBallInOwnPenaltyArea( )
{
  return isInOwnPenaltyArea( getBallPos() );
}

/*! This method returns whether the specified position lies in the own penalty
    area.
    \param pos position which should be checked
    \return bool indicating whether 'pos' lies in own penalty area. */
bool WorldModel::isInOwnPenaltyArea( VecPosition pos )
{
  ObjectT     objFlag = ( getSide() == SIDE_LEFT )
                              ?  OBJECT_FLAG_P_L_C
                              :  OBJECT_FLAG_P_R_C ;
  VecPosition posFlag = SoccerTypes::getGlobalPositionFlag( objFlag, getSide());
  if( pos.getX()         < posFlag.getX() &&
      fabs( pos.getY() ) < PENALTY_AREA_WIDTH/2.0 )
    return true;

  return false;
}

/*! This method returns whether the specified position lies in the opponent
    penalty area.
    \param pos position which should be checked
    \return boolean indicating whether 'pos' lies in opponent penalty area. */
bool WorldModel::isInTheirPenaltyArea( VecPosition pos )
{
  ObjectT     objFlag = ( getSide() == SIDE_LEFT )
                              ?  OBJECT_FLAG_P_R_C
                              :  OBJECT_FLAG_P_L_C ;
  VecPosition posFlag = SoccerTypes::getGlobalPositionFlag( objFlag, getSide());

  if ( pos.getX() > posFlag.getX() &&
       fabs(pos.getY()) < PENALTY_AREA_WIDTH/2.0 )
    return true;

  return false;
}

/*! these three methods returns true if the object is inside defensive or other
  regions*/

bool WorldModel::isInDefenseArea( ObjectT o)
{
  VecPosition posObject = getGlobalPosition(o);

  if (posObject.getX() < -PITCH_LENGTH/5.0)
    return true;
  else
    return false;
}  

bool WorldModel::isInOffenseArea( ObjectT o)
{
  VecPosition posObject = getGlobalPosition(o);

  if (posObject.getX() > PITCH_LENGTH/4.0)
    return true;
  else
    return false;
} 

bool WorldModel::isInMiddleArea( ObjectT o)
{
  

  if (!(isInDefenseArea(o) || isInOffenseArea(o)))
    return true;
  else
    return false;
}

/*! This method determines whether the confidence for 'o' is good. The
    confidence of the object is compared to the player_conf_thr defined in
    PlayerSettings. When the confidence is higher than this value and the object
    does not equal the agent object type true is returned, otherwise false.
    \param o object of which confidence value should be returned
    \return bool indicating whether object information has good confidence. */
bool WorldModel::isConfidenceGood( ObjectT o )
{
  return getConfidence( o ) > PS->getPlayerConfThr() &&
         o != getAgentObjectType();
}

/*! This method determines whether the confidence for 'o' is very good. The
    confidence of the object is compared to the player_high_conf_thr defined in
    PlayerSettings. When the confidence is higher than this value and the object
    does not equal the agent object type true is returned, otherwise false.
    \param o object of which confidence value should be returned
    \return bool indicating whether object information has good confidence. */
bool WorldModel::isConfidenceVeryGood( ObjectT o )
{
  return getConfidence( o ) > PS->getPlayerHighConfThr() &&
         o != getAgentObjectType();
}

/*! This method checks whether the specified object stands onside. This is done
    by comparing the x coordinate of the object to the offside line.
    \return boolean indicating whether 'obj' stands onside. */
bool WorldModel::isOnside( ObjectT obj )
{
  return getGlobalPosition( obj ).getX() < getOffsideX() - 0.5;
}

/*! This method determines whether there stands an opponent in the global
    direction of the specified angle and in distance 'dDist'. An opponent is
    considered to stand in the global direction when the angle difference with
    the specified angle is smaller than 60 degrees.
    \param ang angle of the global direction in which to check opponents
    \param dDist distance in which opponents should be checked
    \return bool indicating wheter an opponent was found. */
bool WorldModel::isOpponentAtAngle( AngDeg ang , double dDist )
{
  VecPosition posAgent   = getAgentGlobalPosition();
  VecPosition posOpp;
  AngDeg      angOpp;
  int         iIndex;

  for( ObjectT o = iterateObjectStart( iIndex, OBJECT_SET_OPPONENTS );
       o != OBJECT_ILLEGAL;
       o = iterateObjectNext ( iIndex, OBJECT_SET_OPPONENTS ) )
  {
    posOpp    = getGlobalPosition( o );
    angOpp    = ( posOpp - posAgent ).getDirection() ;
    if( fabs( angOpp - ang ) < 60 &&
        posAgent.getDistanceTo( posOpp ) < dDist )
      return true;
    else if( fabs( angOpp - ang ) < 120 &&
             posAgent.getDistanceTo( posOpp ) < dDist/2.0 )
      return true;
  }
  iterateObjectDone( iIndex );
  return false;
}
                                 
/* this method returns true if dribbling is good decision. Based on ...
*/           
bool WorldModel::canItoDribble( AngDeg direction, DribbleT speed)
{
  double dist;
  //VecPosition posAgent = getAgentGlobalPosition();
  ObjectT closestOpponent = getClosestInSetTo(OBJECT_SET_OPPONENTS, OBJECT_BALL);
  VecPosition posOpp  = getGlobalPosition(closestOpponent);
  TiredNessT tired = getAgentTiredNess();
  AngDeg dirOpp = (posOpp-getAgentGlobalPosition()).getDirection();

  switch( speed)
  {
    case DRIBBLE_FAST:
      dist = PS->getFastDribbleBlockingDist();
    break;
    case DRIBBLE_SLOW:
      dist = PS->getSlowDribbleBlockingDist();
    break;
    default:
      dist = 1.5;
    break;
  }

  //don't dribble if the target point closer than 3 m to line og opponent's goal
  if( (getBallPos() + VecPosition(dist, direction, POLAR)).getX() > PITCH_LENGTH/2.0 - 4.0)
    return false;

  // don't dribble very close to the own goal
  if( getBallPos().getX() < -PITCH_LENGTH/2.0 + 12.0 &&
       fabs(getBallPos().getY()) < 8.0)
   return false;

 

  // defenders can not dribble on the other half of the pitch
  if(getPlayerNumber() < 6 && getBallPos().getX() > 0 && closestOpponent != getOppGoalieType() &&
     closestOpponent != OBJECT_ILLEGAL )
  {
    return false;
  }
  
  /*else if (tired == TIREDNESS_VERY_BAD )
  {
     cout<<"tired very Bad"<<endl;
     return false;
  }   */

  if (tired == TIREDNESS_VERY_BAD &&
      getBallPos().getDistanceTo(getPosOpponentGoal()) > 25.0 && speed != DRIBBLE_WITHBALL)
  {
     //cout<<"tired Bad"<<endl;
     return false;
  }

  if (closestOpponent != OBJECT_ILLEGAL && getAgentGlobalPosition().getDistanceTo(posOpp) <  PS->getOppVeryCloseThreshold() &&
      fabs(direction - dirOpp) <  105.0)
  {
    //cout<<"too close "<<getPlayerNumber()<<endl;
      return false;
  }

   if (speed == DRIBBLE_WITHBALL)
  {
   // if (closestOpponent != OBJECT_ILLEGAL && getAgentGlobalPosition().getDistanceTo(posOpp) <  PS->getOppCloseThreshold() &&
  //      fabs(direction - dirOpp) <  105.0 )
   // {
   //   return false;
  //  }
     if (getNrInSetInCone(OBJECT_SET_OPPONENTS, 1.0, getBallPos(),
      getBallPos() + VecPosition(dist*2.0, direction, POLAR)) < 1 )
    {
       return true;
    }
    else
    {
      return false;
    }
  }
  if (speed == DRIBBLE_FAST)
  {
    if (getAgentSpeed() < EPSILON)
    {
      return false;
    }
    else if ( fabs(getAgentGlobalBodyAngle()-direction) > 20.0)
    {
      return false;
    }
    else if (getBallPos().getDistanceTo(getPosOpponentGoal()) < 16.0 )
    {
      return false;
    }
    else if (getNrInSetInCone(OBJECT_SET_OPPONENTS, 1.0, getBallPos(),
      getBallPos() + VecPosition(dist*2.0, direction, POLAR)) < 1 )
    {
       return true;
    }
    /*
    else  if( getNrInSetInCone(OBJECT_SET_OPPONENTS, 1.0, getBallPos(),
        getBallPos() + VecPosition(dist*2.0, direction, POLAR)) < 1 &&
        getNrInSetInCone(OBJECT_SET_OPPONENTS, 2.5, getBallPos(),
        getBallPos() + VecPosition(dist*0.75, direction, POLAR)) < 1)
    {
      return true;
    }
    */
    else
    {
    //"don' know\n";
      return false;
    }
  }

  else //slow or with ball
  {
    //if (getBallPos().getDistanceTo(getPosOwnGoal()) < 16.0 )
    //{
    //  return false;
    //}
    //else
    // number of opponents in the largest sector  1
    int sect1 = getNrInSetInCone(OBJECT_SET_OPPONENTS, 1.0, getBallPos(),
                getBallPos() + VecPosition(dist*2.0, direction, POLAR) );

    // the number of opponents  inside sector 1 but outside of sector 2
    int sect2= getNrInSetInCone(OBJECT_SET_OPPONENTS, 1.0, getBallPos(),
                getBallPos() + VecPosition(dist*2.0, direction, POLAR) );


                //getNrInSetInCone(OBJECT_SET_OPPONENTS, 0.7, getBallPos(),
                //getBallPos() + VecPosition(dist*0.8, direction, POLAR));;

    // the number of opponents  inside sector 2 but outside of sector 3
    int sect3= getNrInSetInCone(OBJECT_SET_OPPONENTS, 1.0, getBallPos(),
                getBallPos() + VecPosition(dist*2.0, direction, POLAR));

                //getNrInSetInCone(OBJECT_SET_OPPONENTS, 0.5, getBallPos(),
                //getBallPos() + VecPosition(dist*0.5, direction, POLAR));

    // let's check the sector 1
    if (sect1+sect2 + sect3 < 1 )
    {
      return true;
    }
    else
    {
    //"don' know\n";
      return false;
    }
  }
  return false;
}


bool WorldModel::canItoOutplay( AngDeg direction)
{
  TiredNessT tired = getAgentTiredNess();
  
  if (tired == TIREDNESS_VERY_BAD &&
      getBallPos().getDistanceTo(getPosOpponentGoal()) > 22.0 )
  {
     //cout<<"tired Bad"<<endl;
     return false;
  }
  else if(getBallPos().getDistanceTo(getPosOwnGoal()) < 20.0  )
  {
     return false;
  }
  else
  {
    return true;
  }
  return true;
}


bool  WorldModel::canIGiveLeadingPass( ObjectT tMate, VecPosition *pos )
{
  bool option;
  VecPosition mypos = getAgentGlobalPosition();
  VecPosition pStart = getGlobalPosition(tMate);
  VecPosition leftTop = VecPosition(pStart.getX() + 1.5, pStart.getY() - 5.0);
  VecPosition rightBottom = VecPosition(pStart.getX() + 12.0, pStart.getY() + 5.0);
  // don't check out of the opponent's goal
  if( rightBottom.getX() > PITCH_LENGTH*0.5 )
  {
    rightBottom.setX( PITCH_LENGTH*0.5 );
  }

  Rectangle rect = Rectangle(leftTop, rightBottom);

  
  if( tMate == getAgentObjectType() )
  {
     option = false;
  }
  else if( pStart.getX() < 0.0 )
  {
     option = false;
  }
   
  else if( pStart.getX() < mypos.getX() )
  {
     option = false;
  }
  else if( tMate == OBJECT_ILLEGAL )
  {
     option = false;
  }
  else if( pStart.getX() > getOffsideX() )
  {
     option = false;
  }
  // if the player too far
  else if(pStart.getX() - getBallPos().getX() > 25.0 )
  {
      option = false;
  }
  else if(fabs(pStart.getY() - getBallPos().getY()) > 25.0 )
  {
      option = false;
  }
  else if( getBallPos().getX() > PITCH_LENGTH*0.5 - 22.0 )
  {
      option = false;
  }
  // I check if it is myself and I am tired
  else if( tMate == getAgentObjectType() && getAgentTiredNess() == TIREDNESS_VERY_BAD )
  {
    option = false;
  }
  else if( getNrInSetInRectangle(OBJECT_SET_OPPONENTS, &rect) < 1 )            // was '< 1'  VK
  {                                                                   
    option = true;
    mypos = pStart + VecPosition( 7.0, (getOpponentGoaliePos() - pStart).getDirection(), POLAR);
  }
/*
  else
  { 
    // the stuff below was added by VK
    option = canIGiveThroughPass( tMate, &mypos );   

    if ( not option )
      cout << " through passing attempt ==failed==" << endl;
    else
      cout << " ++through++ pass is possible " << endl;
  }
*/
  *pos = mypos;
  return option;

}


bool WorldModel::canITurn( AngDeg direction)
{
  ObjectT closestOpponent = getClosestInSetTo(OBJECT_SET_OPPONENTS, OBJECT_BALL);
  VecPosition posOpp  = getGlobalPosition(closestOpponent);
  AngDeg dirOpp = (posOpp-getAgentGlobalPosition()).getDirection();
  
  // check if we need to turn
  if (fabs(getAgentGlobalBodyAngle() - direction) < 105.0) // magic number
    return false;
  // too close to own goal 
  if (getBallPos().getDistanceTo(getPosOwnGoal()) < 16.0 )
  {
     return false;
  }
  // an opponent is too close
  if (getAgentGlobalPosition().getDistanceTo(posOpp) <  PS->getOppVeryCloseThreshold() &&
      fabs(direction - dirOpp) <  105.0)
  {
    return false;
  }
  
  return true;
}

/* this method returns best direction for dribble inside "freedom" in "direction"
 direction. Chosen as direction to point between two closest opponents.
 */

AngDeg  WorldModel::getBestDirForDribble( AngDeg direction, AngDeg freedom)
{


  return (getPointBetweenTwoClosestOpponentsInCone(getBallPos(),
         getBallPos()+VecPosition(1.0, direction-freedom/2.0, POLAR),
         getBallPos()+VecPosition(1.0, direction+freedom/2.0, POLAR),
         PS->getFastDribbleBlockingDist()*2.0)-getBallPos()).getDirection();
  
}

/* this method returns point for direct pass. At first we have to predict position of
the partner, take in consideration a closest opponent and the body direction of the
partner.
*/
VecPosition WorldModel::getBestPointForDirectPass  ( ObjectT o  )
{
  VecPosition posToPass, posPredT, posPredO;
  VecPosition posBall = getBallPos();
  AngDeg bodyDir = getGlobalBodyAngle( o );
  AngDeg ballDir, oppDir, ang;
  double dist;

  //predict the teammate position assuming that it will take for him at least 3 cycles to react
  posPredT = predictPlayerPos(3, o);
  //find the closes opponent to this point
  ObjectT opponent = getClosestInSetTo(OBJECT_SET_OPPONENTS, posPredT, &dist, PS->getPlayerConfThr());
  //predict the opponent position after the same time
  posPredT = predictPlayerPos(3, opponent);


  // if an opponent is far enough, consider prediction and body dir only
  if ( (dist > 3.0 && dist > posBall.getDistanceTo(posPredT)/3.0 )|| opponent == OBJECT_ILLEGAL)  //magic number
  {
    dist =  posBall.getDistanceTo(posPredT)/7.0;
    if(dist > 1.0 )
    {
      dist = 1.0;
    }
    posToPass = posPredT +VecPosition(SS->getPlayerSize(), bodyDir, POLAR) ; // magic number
  }
  else // calculate point furhter from the opponent
  {
    dist =  posBall.getDistanceTo(posPredT)/7.0;
    if(dist > 1.0 )
    {
      dist = 1.0;
    }
    else if(dist < SS->getPlayerSize()/2.0+SS->getBallSize()/2.+1.0)
    {
      dist = SS->getPlayerSize()/2.0+SS->getBallSize()/2.+1.0;
    }
    ballDir = (posBall - posPredT).getDirection();
    oppDir  = (posPredO - posPredT).getDirection();
    ang = VecPosition::normalizeAngle( fabs(ballDir - oppDir) );
    // if the opponent is behind
    if ( ang > 135.0 )
    {
      posToPass = posPredT + VecPosition(dist, ballDir, POLAR);
    }
    //the opponent is on side
    else if( ang > 45.0 )
    {
      ang = VecPosition::normalizeAngle( oppDir - 180.0 );
      posToPass = posPredT + VecPosition(dist, ang, POLAR);
    }
    else if( oppDir > ballDir )
    {
       ang = VecPosition::normalizeAngle( ballDir - 90.0 );
       posToPass = posPredT + VecPosition(dist, ang, POLAR);
    }
    else
    {
      ang = VecPosition::normalizeAngle( ballDir + 90.0 );
      posToPass = posPredT + VecPosition(dist, ang, POLAR);
    }

  }

  return posToPass;
}

bool  WorldModel::canIHideBall()
{
  int numOpp = getNrInSetInCircle(OBJECT_SET_OPPONENTS, Circle(getAgentGlobalPosition(), 3.0));
  double dDist;
  ObjectT closestOpponent = getClosestRelativeInSet(OBJECT_SET_OPPONENTS, &dDist);
  
  if ( dDist < SS->getPlayerSize() + SS->getBallSize()   +  SS->getKickableMargin() )
  {
    return false;
  }
  else if( numOpp > 1)
  {
    return false;
  }
  else if( getBallPos().getX() < 0.0)
  {
    return false;
  }
  else
  {
    return true;
  }
}
/*! This method returns the inverse confidence, i.e. the time that belongs
    to the specified confidence. This can be used to determine
    the time the object was last seen when the confidence is given. Herefore
    the current time is used.
    \param dConf confidence
    \return server cycle the object was last seen. */
Time WorldModel::getTimeFromConfidence( double dConf )
{
  return getCurrentTime()-(int)((1.00-dConf)*100);
}

/*! This method returns the x coordinate of the offside line using the known
    information in the WorldModel. If a player moves beyond this line, he
    stands offside. First the opponent with the second highest x coordinate is
    located, then the maximum of this x coordinate and the ball x coordinate
    is returned.
    \param bIncludeComm boolean indicating whether communicated offside line
          should also be included. 
    \return x coordinate of the offside line. */
double WorldModel::getOffsideX( bool bIncludeComm )
{
  double  dHighestX = 0.0;
  double  dSecondX  = 0.0;
  double  x;
  ObjectT o;
  for( int i = 0; i < MAX_OPPONENTS ; i ++ )
  {
    o = Opponents[i].getType();
    if( isConfidenceVeryGood( o ) )
    {
      x = Opponents[i].getGlobalPosition().getX();
      if( x > dHighestX )         // if larger x than highest
      {
        dSecondX  = dHighestX;    // make second the previous highest
        dHighestX = x;            // and this the new one
      }
      else if( x > dSecondX )     // if smaller than highest and larger than 2nd
        dSecondX = x;             // make it the second
    }
  }
  // if highest x is outside pen_area, it cannot be the goalie (unless playing
  // Portugal ;-) ), so assume goalie is just not seen
  if( dHighestX < PENALTY_X && getOppGoalieType() == OBJECT_ILLEGAL )
    dSecondX = dHighestX;
    
  x = getBallPos().getX();
  x = max( x, dSecondX );

  return x;
}

/*! This method returns the outer position on the field given a position 'pos'
    and a global angle 'ang'. The outer position is defined as the point on
    the field where the line created from this position and angle crosses
    either a side line, goal line or penalty line. To be on the safe side a
    small value is specified, which denotes the distance from the side line
    that should be returned.
    \param pos position on the field from which outer position should be
    calculated
    \param ang global angle which denotes the global direction in pos
    \param dDist distance from line
    \param bWithPenalty boolean denoting whether penalty area should be taken
                        into account (if false only goal line and side line
                        are used.
    \return position denoting the outer position on the field */
VecPosition WorldModel::getOuterPositionInField( VecPosition pos, AngDeg ang,
                                  double dDist, bool bWithPenalty )
{
  VecPosition posShoot;

  // make shooting line using position and desired direction
  Line lineObj     = Line::makeLineFromPositionAndAngle( pos, ang );

  // get intersection point between the created line and goal line
  Line lineLength  = Line::makeLineFromPositionAndAngle(
                            VecPosition( PITCH_LENGTH/2.0 - dDist, 0.0 ), 90 );
  posShoot         = lineObj.getIntersection( lineLength );

  // check whether it first crosses the penalty line
  Line linePenalty = Line::makeLineFromPositionAndAngle(
                            VecPosition( PENALTY_X - dDist, 0.0 ), 90.0 );
  double dPenaltyY = lineObj.getIntersection(linePenalty).getY();

  if( bWithPenalty && fabs(dPenaltyY) < PENALTY_AREA_WIDTH/2.0 )
  {
    if( fabs(dPenaltyY) < PENALTY_AREA_WIDTH/2.0 - 5.0 ||   // crosses inside
        fabs(posShoot.getY()) <  PENALTY_AREA_WIDTH/2.0 )   // or ends inside
      posShoot = lineObj.getIntersection( linePenalty );
  }

  // check where it crosses the side line
  Line lineSide = ( ang < 0 )
     ? Line::makeLineFromPositionAndAngle(
                           VecPosition( 0.0, - PITCH_WIDTH/2.0 + dDist ), 0.0 )
     : Line::makeLineFromPositionAndAngle(
                           VecPosition( 0.0, + PITCH_WIDTH/2.0 - dDist ), 0.0 );

  if( fabs(posShoot.getY()) > PITCH_WIDTH/2.0 - dDist )
    posShoot = lineObj.getIntersection( lineSide );

  return posShoot;
}



VecPosition WorldModel::getPointBetweenTwoClosestOpponentsInCone (VecPosition pStart,
                VecPosition pLeft, VecPosition pRight, double dist)
{
  VecPosition opp1;
  VecPosition opp2;
  
  double      dMinMag         = 1000.0;
  double      dMinMag1        = 1000.0;
  VecPosition v = VecPosition(-1000.0,-1000.0);
  AngDeg angle1 = (pLeft-pStart).getDirection();
  AngDeg angle2 = (pRight-pStart).getDirection();
    
  AngDeg angle3 = 180.0;;
  AngDeg angle4 = 180.0;
  
  angle1 = VecPosition::normalizeAngle(angle1);
  angle2 = VecPosition::normalizeAngle(angle2);
  if (angle1 > angle2)
  {
    AngDeg temp1 = angle2;
    angle2 = angle1;
    angle1 = temp1;
  }
  //cout<<"angles 1"<< angle1<<" 2nd "<<  angle2<<endl;
  int         iIndex;

  double dConfThr      = PS->getPlayerConfThr();
  for( ObjectT o = iterateObjectStart( iIndex, OBJECT_SET_OPPONENTS, dConfThr );
       o != OBJECT_ILLEGAL;
       o = iterateObjectNext ( iIndex, OBJECT_SET_OPPONENTS, dConfThr ) )
  {
    v = getGlobalPosition( o ) - pStart;
    if (VecPosition::normalizeAngle(v.getDirection()) < angle1 ||
        VecPosition::normalizeAngle(v.getDirection()) > angle2 || v.getMagnitude() > dist )
        continue;
    if( v.getMagnitude() < dMinMag )
    {
      dMinMag1        = dMinMag;
      opp2              = opp1;
      angle4          = angle3;
      dMinMag         = v.getMagnitude();
      opp1   = v;
      angle3        =  VecPosition::normalizeAngle(v.getDirection());
    }
    else if ( v.getMagnitude() < dMinMag1 )
    {
       dMinMag1         = v.getMagnitude();
      opp2   = v;
      angle4        =  VecPosition::normalizeAngle(v.getDirection());
    }
  }
  iterateObjectDone( iIndex );

 // if (MinMag > 900.0)
   // return pStart + VecPosition(5.0, angle2+(angle1-angle2)/2, CS_GGGG);
  if(dMinMag1 < 1000.0)
  {
    //cout<<"2 opponents\n";
    
    if (angle3 > angle4)
    {
      AngDeg temp = angle4;
      angle4 = angle3;
      angle3 = temp;
    }
    //cout<<angle1<<" "<<angle3<<" "<<angle4<<" "<<angle2<<endl;
    if (angle3-angle1 > angle4 -angle3 &&  angle3-angle1 > angle2 -angle4)
       v = pStart + VecPosition(5.0, angle1+(angle3-angle1)/2.0-5.0, POLAR);
    else if (angle4-angle3 >  angle2 -angle4)
       v = pStart + VecPosition(5.0, angle3+(angle4-angle3)/2.0, POLAR);
    else
       v = pStart + VecPosition(5.0, angle4+(angle2-angle3)/2.0+5.0, POLAR);   
  }
  else if(dMinMag < 1000.0)
  {
    //cout<<"1 opponent\n";
    //cout<<angle1<<" "<<angle3<<" "<<angle2<<endl;
    if (angle3-angle1 >  angle2 - angle3)
    {
       v = pStart + VecPosition(5.0, angle1+(angle3-angle1)/2.0-5.0, POLAR);
    }
    else
    {
       v = pStart + VecPosition(5.0, angle3+(angle2-angle3)/2.0+5.0, POLAR);
    }
  }
  else
  {
   // cout<<"no opponents\n";
    v = pStart + VecPosition(5.0, angle1+(angle2-angle1)/2.0, POLAR);
  }
  //cout<<"shoot dir "<<(v-pStart).getDirection()<<endl;
  return v;
}


VecPosition WorldModel::getPointBetweenTwoClosestOpponentsInRect (VecPosition pStart,
                        VecPosition pLeftTop, VecPosition pRightBot)
{
  VecPosition opp1;
  VecPosition opp2;

  VecPosition pLeft = VecPosition( (pLeftTop.getX() + pRightBot.getX())*0.5, pLeftTop.getY() );
  VecPosition pRight = VecPosition( (pLeftTop.getX() + pRightBot.getX())*0.5, pRightBot.getY() );

  double      dMinMag         = 1000.0;
  double      dMinMag1        = 1000.0;
  VecPosition v = VecPosition(-1000.0,-1000.0);

  Rectangle rect =Rectangle(pLeftTop, pRightBot) ;
  
  
  int         iIndex;

  double dConfThr      = PS->getPlayerConfThr();
  for( ObjectT o = iterateObjectStart( iIndex, OBJECT_SET_OPPONENTS, dConfThr );
       o != OBJECT_ILLEGAL;
       o = iterateObjectNext ( iIndex, OBJECT_SET_OPPONENTS, dConfThr ) )
  {
    v = getGlobalPosition( o );
    if (!rect.isInside(v) )
    {
        continue;
    }
    if( (v-pStart).getMagnitude() < dMinMag )
    {
      dMinMag1        = dMinMag;
      opp2              = opp1;
      dMinMag         = (v-pStart).getMagnitude();
      opp1   = v;
    }
    else if ( v.getMagnitude() < dMinMag1 )
    {
      dMinMag1         = (v-pStart).getMagnitude();
      opp2   = v;
      
    }
  }
  iterateObjectDone( iIndex );

 
  if(dMinMag1 < 1000.0)
  {
    //cout<<"2 opponents\n";
    if (opp1.getY() > opp2.getY()) // we will count from left to right
    {
      VecPosition temp = opp1;
      opp1 = opp2;
      opp2 = temp;
    }
    
    if ( (pLeft-opp1).getMagnitude() > (opp1 - opp2).getMagnitude() &&
        (pLeft-opp1).getMagnitude() > (opp2 - pRight).getMagnitude())
       v = pLeft.getVecPositionOnLineFraction(opp1, 0.5);
    else if ((opp1 - opp2).getMagnitude()  >  (opp2 - pRight).getMagnitude())
       v = opp1.getVecPositionOnLineFraction(opp2, 0.5);
    else
       v = pRight.getVecPositionOnLineFraction(opp2, 0.5);
  }
  else if(dMinMag < 1000.0)
  {
    //cout<<"1 opponent\n";
    
    if ( (pLeft-opp1).getMagnitude() > (opp1 - pRight).getMagnitude())
    {
       v = pLeft.getVecPositionOnLineFraction(opp1, 0.5);
    }
    else
    {
       v = pRight.getVecPositionOnLineFraction(opp1, 0.5);
    }
  }
  else
  {
    //cout<<"no opponents\n";
    v = pLeft.getVecPositionOnLineFraction(pRight, 0.5);
  }
  //cout<<"shoot dir "<<(v-pStart).getDirection()<<endl;
  return v;
}



/*! This method determines the (global) direction which has the largest
    angle between the opponents and is located in the interval angMin..
    angMax.
    \param origin of which the angles angMin and angMax are based on.
    \param angMin minimal global direction that should be returned
    \param angMax maximal global direction that should be returned
    \param angLargest will contain the size of the largest angle of the
           direction that is returned
    \param dDist only opponents with relative distance smaller than this value
           will be taken into account.
    \return global direction with the largest angle between opponents */
AngDeg WorldModel::getDirectionOfWidestAngle(VecPosition posOrg, AngDeg angMin,
                               AngDeg angMax, AngDeg *angLargest, double dDist)
{
  list<double> v;
  list<double> v2;
  double       temp;
  int          iIndex;
  double       dConf  = PS->getPlayerConfThr();

  // add all angles of all the opponents to the list v
  for( ObjectT o = iterateObjectStart( iIndex, OBJECT_SET_OPPONENTS, dConf );
       o != OBJECT_ILLEGAL;
       o = iterateObjectNext ( iIndex, OBJECT_SET_OPPONENTS, dConf ) )
  {
    if( getRelativeDistance( o ) < dDist )
      v.push_back( (getGlobalPosition(o)-posOrg).getDirection());
  }
  iterateObjectDone( iIndex );
  v.sort();

  // if goalkeeper is spotted and he is located within the range that we want
  // to shoot at, make sure the angle with the goalkeeper is large enough, since
  // he has better intercepting capabilities than the normal players
  
  ObjectT     objGoalie = getOppGoalieType();
  VecPosition posGoalie = getGlobalPosition( objGoalie );
  AngDeg      angGoalie;
  
  if( objGoalie != OBJECT_ILLEGAL && posOrg.getX() > PITCH_LENGTH/4.0 &&
      posOrg.getDistanceTo( posGoalie ) < dDist )
  {
    angGoalie = ( posGoalie - posOrg ).getDirection();  
    Log.log( 560, "direction_widest_angle: min %f max %f angGoalie %f", 
                                                  angMin, angMax, angGoalie );    
    if( posOrg.getY() > 0 ) // right side of the field
    {
      angGoalie = VecPosition::normalizeAngle( angGoalie - 33 );
      angMax    = max( angMin, min( angGoalie, angMax ) );
    }
    else
    {
      angGoalie = VecPosition::normalizeAngle( angGoalie + 33 );
      angMin    = min( angMax, max( angMin, angGoalie ) );      
    }
    Log.log( 560, "direction_widest_angle after: %f %f", angMin, angMax );
  }
       

  // Create new list with only opponents from interval [angMin..angMax].
  // Note that opponents outside angMin and angMax can have an influence
  // on the largest angle between the opponents, so they should be accounted
  // for. To this end, a projection is defined in both angMin and angMax.
  // The opponent with the smallest global angle difference a to angMin
  // (either inside or outside the interval [angMin..angMax]) is determined
  // and an extra angle angMin - a is added to the list. The situation for
  // angMax is analogous.

  double absMin     = 1000;
  double absMax     = 1000;
  double angProjMin = angMin;
  double angProjMax = angMax;
  double array[MAX_OPPONENTS+2];

  while( v.size() > 0 )
  {
    if( fabs( v.front() - angMin ) < absMin )      // opp near angMin
    {
      absMin     = fabs( v.front() - angMin ) ;    // make angMin wider
      angProjMin = angMin - absMin;                // to take him into account
    }
    if( fabs( v.front() - angMax ) < absMax )      // opp near angMax
    {
      absMax     = fabs( v.front() - angMax ) ;    // make angMax wider
      angProjMax = angMax + absMax;                // to take him into account
    }
    if( v.front() > angMin && v.front() < angMax ) // opp in range
      v2.push_back( v.front() );                   // add him
    v.pop_front();
  }

  // make all angles relative to angProjMin which has angle 0 and set them in
  // the range 0..360, where the range -180..0 is moved to 180..360. Do this by
  // adding 360 and then subtracting 360 if value is larger than 360.
  v.push_back( 0 );
  while( v2.size() > 0 )  // for all the opponents
  {
    temp = VecPosition::normalizeAngle(v2.front()-angProjMin)+360.0;
    if( temp > 360 )
      temp -= 360;
    v.push_back( temp );
    v2.pop_front();
  }
  // add max projection.
  temp = VecPosition::normalizeAngle(angProjMax-angProjMin)+360.0;
  if( temp > 360 )
    temp -= 360;

  v.push_back( temp );

  // sort the list
  v.sort();

  // put all the values in an array
  int i = 0;
  while( v.size() > 0 )
  {
    array[i++] = v.front();
    v.pop_front();
  }

  // find the largest angle and determine the associated midpoint direction
  double dLargest = -1000;
  double d;
  double ang      = UnknownAngleValue;
  for( int j = 0; j < i - 1 ; j ++ )
  {
    d = VecPosition::normalizeAngle(( array[j+1] - array[j] )/2.0);
    if( d > dLargest )
    {
      ang = angProjMin + array[j] + d;
      ang = VecPosition::normalizeAngle( ang );
      dLargest = d;
    }
  }

  if( ang == UnknownAngleValue ) // no angle found -> get angle in between
  {
    ang = getBisectorTwoAngles( angMin, angMax );
    if( angLargest != NULL )
      *angLargest = 360;
  }
  else if( angLargest != NULL )
    *angLargest = dLargest;

  return ang;
}


/*! This method determine the strategic position for the specified player. This 
    is done using the Formations class. In this class all information
    about the current formation, player number in formation and otheic
    values are stored. The strategic position is based on the position of
    the ball. If the confidence in the position of the ball is lower than the
    threshold defined in PlayerSettings, it is assumed that the ball is at
    position (0,0).
    \param iPlayer role in formation for which strategic position should be 
      returnd. With default value is -1 it is assumed that the strategic
      position of the agent itself should be returned. 
    \return VecPosition strategic position for player 'iPlayer' */
VecPosition WorldModel::getStrategicPosition( int iPlayer )
{
  VecPosition pos, pos1; //, pos2;
  double dd = 0;
  double dd1 = 0;
  double  dMinX;

  // -1 is default -> get player number in formation
  if( iPlayer == -1 )
  {
    iPlayer = formations->getPlayerInFormation();
  }

  // get maximal allowed x coordinate, this is offside x coordinate
  //double dMaxX = getOffsideX() - 3.0;
  double dMaxX = getOffsideX() - 0.5;

  // after standing offside we are not allowed to move for ball
  // with a goal kick of them we are not allowed to move into penalty area
  VecPosition posBall = predictPosAfterNrCycles(OBJECT_BALL, 2);
  if( isOffsideUs( ) )
    dMaxX = min( dMaxX, getBallPos().getX() - 3.0 );
  if( isGoalKickThem() || (isFreeKickThem() && posBall.getX() > PENALTY_X )  )
    dMaxX = min( dMaxX, PENALTY_X - 1.0 );
  if( isBeforeKickOff() )
    dMaxX = min( dMaxX, -2.0 );

  // change the ball position on which strategic position is based depending on
  // the different deadball situation and thus the expected movement of the ball
  //VecPosition posBall = getBallPos();

  //if( getTimeSinceLastCatch(  ) < PS->getCyclesCatchWait() + 15  )
  if( getTimeSinceLastCatch(  ) < 48 + 15  )
    posBall.setX( -PITCH_LENGTH/2 + 32.0 );
  else if(isGoalKickUs() )
    posBall.setX( -PITCH_LENGTH/2 + 22.0 );
  else if( getConfidence( OBJECT_BALL ) < PS->getBallConfThr() )
    posBall.setVecPosition( 0.0, 0.0 );
  else if( isGoalKickThem() ||
           ( isFreeKickThem() && posBall.getX() > PENALTY_X ) )
    posBall.setX( PENALTY_X - 10.0 );

  VecPosition posTarget = getTargetArea();
  VecPosition posPlayer = getAgentGlobalPosition();
  PossesionT ownBall = getTeamPossesion();
  // get the strategic position
  pos = formations->getStrategicPosition1( iPlayer, posBall, posTarget,
                                           posPlayer, ownBall, dMaxX, isInOffenseArea(OBJECT_BALL) );
  //pos = formations->getStrategicPosition( iPlayer, posBall, dMaxX );
  if (getTimeSinceLastCatch()< PS->getCyclesCatchWait() + 35  || isGoalKickUs())
  {
    pos = VecPosition(pos.getX(), pos.getY()*1.2 );
    //if (isGoalKickUs())
     // pos = VecPosition(pos.getX()-8.0, pos.getY());
  }
  // special positioning for the attack

  // defind starting Y coordinate for positioning
  dd1 = pos.getY();
  if( dd1 < -PITCH_LENGTH*0.5 +12.5 ) // if too left
  {
    dd1 = -PITCH_LENGTH*0.5 +12.5;
  }

  if( dd1 > PITCH_LENGTH*0.5 - 12.5 ) // if too right
  {
    dd1 = PITCH_LENGTH*0.5 - 12.5;
  }

  if( getBallPos().getX() > 0.0 &&
      ( formations->getPlayerType() == PT_ATTACKER_WING ||
        formations->getPlayerType() == PT_ATTACKER        ) )    // attackers
  {
      pos = VecPosition( dMaxX, pos.getY() );
  }
  else if( formations->getPlayerType() == PT_ATTACKER_WING ||
           formations->getPlayerType() == PT_ATTACKER )          //the ball is in the defense area
  {
   // dd = -PITCH_LENGTH/2.0 + 26.0;
    dd = 0.0;
    pos.setX( dMaxX - ( dd - posBall.getX() )/3.0 );
  }


  //  special positioning for the defense
  //else if ( isInDefenseArea(OBJECT_BALL) && isBallInOurPossesion()!=POSSESION_US)
  else if ( posBall.getX() < -PITCH_LENGTH/2.0 + 20.0 && isBallInOurPossesion()!=POSSESION_US)
  {
    dd = -PITCH_LENGTH/2.0 + 11.0;
    switch(iPlayer)
    {
      case 1:     // player #2
        pos = VecPosition(-PITCH_LENGTH/2.0+11.0+(posBall.getX()-dd)/2.0, pos.getY());//*0.85 );
        break;
      case 2:     //player #3
        pos = VecPosition(-PITCH_LENGTH/2.0+11.0+(posBall.getX()-dd)/2.0,pos.getY());//*0.85 );
        break;
      case 3:     // player #4
        pos = VecPosition(-PITCH_LENGTH/2.0+11.0+(posBall.getX()-dd)/2.0, pos.getY());//*0.85);
        break;
      case 4:     // player #5
        pos = VecPosition(-PITCH_LENGTH/2.0+11.0+(posBall.getX()-dd)/2.0, pos.getY());//*0.85 );
        break;

    }
  }

  // special positioning for midfielders
  if ( formations->getPlayerType() == PT_MIDFIELDER_WING)
  {
    // find X position for a central  defender
    pos1 = formations->getStrategicPosition1( 2, posBall, posTarget,
                                           posPlayer, ownBall, dMaxX, isInOffenseArea(OBJECT_BALL) );
    dMinX = pos1.getX();

    pos.setX( (dMinX+dMaxX)*0.5 + (posBall.getX() -  (dMinX+dMaxX)*0.5)*0.5);
    if(isInDefenseArea(OBJECT_BALL))
    {
       pos.setX( (dMinX+dMaxX)*0.5 + (posBall.getX() -  (dMinX+dMaxX)*0.5)*0.5);
       pos.setY(pos.getY()*0.85);
    }

  }


  else if( formations->getPlayerType() == PT_MIDFIELDER_CENTER )
  {
    // find X position for a central  defender
    pos1 = formations->getStrategicPosition1( 2, posBall, posTarget,
                                           posPlayer, ownBall, dMaxX, isInOffenseArea(OBJECT_BALL) );
    dMinX = pos1.getX();

    pos.setX( (dMinX+dMaxX)*0.5 + (posBall.getX() -  (dMinX+dMaxX)*0.5)*0.3);

  }

  if( formations->getPlayerType() == PT_MIDFIELDER_WING ||
      formations->getPlayerType() == PT_MIDFIELDER_CENTER)
  {
       pos = getPointBetweenTwoClosestOpponentsInRect(
                   pos,
                   VecPosition( pos.getX() - 3.0, dd1 - 12.5 ),
                   VecPosition( pos.getX() + 3.0, dd1 + 12.5 )
                                                    );
  }
  
  if( ( (!isInDefenseArea(OBJECT_BALL) || isDeadBallThem() ) && ( iPlayer > 7) ) ||
       isDeadBallUs() ) // the ball is on the opponent's half of the field
  {

     
     pos = getPointBetweenTwoClosestOpponentsInRect(
                   pos,
                   VecPosition( pos.getX() - 3.0, dd1 - 12.5 ),
                   VecPosition( pos.getX() + 3.0, dd1 + 12.5 )
                                                    );

  }

  return pos;
}


/*This method calculates a position for player who has to kick the ball in
when it is a dead ball us.
*/

VecPosition WorldModel::getPosForDeadBall()
{
    VecPosition posTarget;
    VecPosition posBall = getBallPos();

    if( posBall.getDistanceTo(VecPosition(0,0)) > PITCH_WIDTH/2.0-1.0 )
    {
      posTarget = posBall+VecPosition(0.5, posBall.getDirection(), POLAR);
    }
    else
    {
      posTarget = VecPosition(posBall.getX()-0.5, posBall.getY());
    }

    return posTarget;
}


/*! The actual power with which the ball is kicked depends on the
    relative location of the ball to the player. The kick is more
    powerful when the ball is very close to and in front of the player.
    The actual kickpowerrate with which the power of the kick command
    is multiplied is equal to<BR>
    KickPowerRate*(1 - 0.25*DirDiff / 180 -
    0.25*(DistBall-PlayerSize-BallSize)/KickableMargin)<BR>
    with DirDiff = global angle of the ball rel to the body dir agent<BR>
         DistBall = the distance from the center of the player to the ball<BR>
    See soccermanual
    \return the actual kick power rate with which power is multiplied  */
double WorldModel::getActualKickPowerRate( )
{
 // true indicates that relative angle to body should be returned
 double dir_diff      = fabs( getRelativeAngle( OBJECT_BALL, true ) );
 double dist          = getRelativeDistance( OBJECT_BALL ) -
                        SS->getPlayerSize( ) - SS->getBallSize( );
 return SS->getKickPowerRate() *
           ( 1 - 0.25 * dir_diff/180.0 - 0.25 * dist / SS->getKickableMargin());
}

/*! The actual power with which the ball must be kicked depends on the
    relative location of the ball to the player. The kick is more
    powerful when the ball is very close to and in front of the player.
    The actual power with which the ball must be kicked is equal to<BR>
    Speed / KickPowerRate*(1 - 0.25*DirDiff / 180 -
    0.25*(DistBall-PlayerSize-BallSize)/KickableMargin)<BR>
    with DirDiff = global angle of the ball rel to the body dir agent<BR>
         DistBall = the distance from the center of the player to the ball<BR>
    See soccermanual for further information.
    This method receives a speed vector which the ball should have after the
    kick command and calculates the power for the kick command to reach this.
    This value can be higher than is possible to shoot!
    \param dDesiredSpeed the desired speed after the kick command
    \return the actual power for kick command to get dDesiredSpeed */
double WorldModel::getKickPowerForSpeed( double dDesiredSpeed )
{
  // acceleration after kick is calculated by power * eff_kick_power_rate
  // so actual kick power is acceleration / eff_kick_power_rate
  return dDesiredSpeed / getActualKickPowerRate( );
}


/*! This method determines the power with which the ball must be kicked in
    order to travel a given distance and still have a speed after that
    distance
    \param dDistance distance ball should travel
    \param dEndSpeed speed ball should have at target position
    \return power value for kick command */
double WorldModel::getKickSpeedToTravel( double dDistance, double dEndSpeed )
{
  // if endspeed is zero we have an infinite series and return the first term
  // that corresponds to the distance that has to be travelled.
  if( dEndSpeed < 0.0001  )
    return Geometry::getFirstInfGeomSeries(dDistance, SS->getBallDecay() );

  // use geometric series to calculate number of steps and with that the
  // velocity to give to the ball, we start at endspeed and therefore use
  // the inverse of the ball decay (r).
  // s = a + a*r + .. a*r^n since we calculated from endspeed (a) to
  // firstspeed, firstspeed equals a*r^n = endspeed*r^nr_steps
  double dNrSteps = Geometry::getLengthGeomSeries( dEndSpeed,
                                  1.0/SS->getBallDecay( ), dDistance );
  return getFirstSpeedFromEndSpeed( dEndSpeed, (int)rint(dNrSteps) ) ;
}

/*! This method returns the speed that has to be given to the ball when
    it should have an endspeed of 'dEndSpeed' after 'dCycles' number of cycles.
    This can be calculated using a geometric series.
    \param dEndSpeed desired end speed for the ball
    \param dCycles nr of cycles after which ball should have speed 'dEndSpeed'
    \return initial speed given to the ball to have speed 'dEndSpeed' after
    'dCycles' cycles. */
double WorldModel::getFirstSpeedFromEndSpeed( double dEndSpeed, double dCycles )
{
  // geometric series: s = a + a*r^1 + .. + a*r^n, now given endspeed = a*r^n ->
  // endspeed = firstspeed * ratio ^ length ->
  // firstpeed = endspeed * ( 1 / ratio ) ^ length
  return dEndSpeed * pow( 1 / SS->getBallDecay(), dCycles );
}

/*! This method returns the speed that has to be given to the ball when
    it should have travelled a distance 'dDist' after 'dCycles' number of
    cycles. This can be calculated using a geometric series.
    \param dDist distance the ball has to travel
    \param dCycles nr of cycles after which ball should have travelled 'dDist'
    \return initial speed for the ball to travel 'dDist' in 'dCycles' cycles */
double WorldModel::getFirstSpeedFromDist( double dDist, double dCycles )
{
  return Geometry::getFirstGeomSeries( dDist, SS->getBallDecay(), dCycles);
}

/*! This method returns the speed the ball will have after 'dCycles' cycles
    when it is given an initial speed of 'dFirstSpeed'.
    This can be calculated using a geometric series.
    \param dFirstSpeed given speed to the ball
    \param dCycles nr of cycles after which ball speed should be determined
    \return speed of the ball after 'dCycles' server cycles */
double WorldModel::getEndSpeedFromFirstSpeed(double dFirstSpeed, double dCycles)
{
  // geometric series: s = a + a*r^1 + .. + a*r^n, with firstspeed = a ->
  // endspeed = firstspeed * ratio ^ length ;
  return dFirstSpeed * pow( SS->getBallDecay(), dCycles );
}

/*! This method determines the angle that should be sent to the soccerserver when
    the player wants to turn angDesiredAngle. This value depends on the current
    velocity and the inertia moment of the player
    \param angDesiredAngle angle that player wants to turn
    \param dSpeed current speed of the player
    \return angle that can be sent with turn command */
AngDeg WorldModel::getAngleForTurn( AngDeg angDesiredAngle, double dSpeed )
{
  AngDeg a = angDesiredAngle * (1.0 + SS->getInertiaMoment() * dSpeed );
  if( a > SS->getMaxMoment() )
    return SS->getMaxMoment() ;
  else if ( a < SS->getMinMoment() )
    return SS->getMinMoment() ;
  else
    return a;
}

/*! This method determines the actual angle that is used when 'angTurn' is
    sent to the SoccerServer. This value depends on the current velocity and
    the inertia moment of the player
    \param angAngleForSend angle send with turn command
    \param dSpeed current speed of the player
    \return actual angle that player is turned */
AngDeg WorldModel::getActualTurnAngle( AngDeg angTurn, double dSpeed )
{
  return angTurn / (1.0 + SS->getInertiaMoment() * dSpeed );
}

/*! This method determines the optimal dash power to mantain an optimal speed
    When the current speed is too high and the distance is very small, a
    negative dash is performed. Otherwise the difference with the maximal speed
    is determined and the dash power rate is set to compensate for this
    difference.
    \param posRelTo relative point to which we want to dash
    \param angBody body angle of the agent
    \param vel current velocity of the agent
    \param dEffort current effort of the player
    \return dash power that should be sent with dash command */
double WorldModel::getPowerForDash( VecPosition posRelTo, AngDeg angBody,
              VecPosition vel, double dEffort )
{
  // the acceleration desired is the x-direction to the relative position we
  // we want to move to. If point lies far away, we dash maximal. Furthermore
  // we subtract the x contribution of the velocity because it is not necessary
  // to dash maximal.
  double dAcc = posRelTo.rotate(-angBody).getX();  // get distance in direction
  if( dAcc > SS->getPlayerSpeedMax() )             // if too far away
    dAcc = SS->getPlayerSpeedMax();                // set maximum speed
  dAcc -= vel.rotate(-angBody).getX();             // subtract current velocity

  // acceleration = dash_power * dash_power_rate * effort ->
  // dash_power = acceleration / (dash_power_rate * effort )
  double dDashPower = dAcc/(SS->getDashPowerRate() * dEffort );
  if( dDashPower > SS->getMaxPower() )
    return SS->getMaxPower();
  else if( dDashPower < SS->getMinPower() )
    return SS->getMinPower();
  else
    return dDashPower;
}

/*! This method determines if the ball is within the specified region

    \return bool true if the ball is in the region

*/
bool WorldModel::isBallWithinRegion(RegionT region)
{
  return isObjectWithinRegion(OBJECT_BALL, region);
}

/*! This method determines if the specified object is within the
    specified region.

    \return bool true if the object is in the region

*/
bool WorldModel::isObjectWithinRegion(ObjectT object, RegionT region)
{
  return getRegionPtrFromType(region)->isInRegion(getGlobalPosition(object));
}

/*! This method determines if a specified position lies within a
    specified region.

    \return bool true if the position is within the region

*/
bool WorldModel::isWithinRegion(const VecPosition& pos, RegionT region)
{
  return getRegionPtrFromType(region)->isInRegion(pos);
}

/*! This method determines if the object specified is within one
    of the regions bordering the side of the field, in the offensive
    zone. Specifically regions 3, 4, and 5.

    \return bool true if the object is within a side region

*/
bool WorldModel::isObjectAlongSide(ObjectT object)
{
  if (object == OBJECT_ILLEGAL)
    return false;
  VecPosition pos = getGlobalPosition(object);
  //is the object within regions 3,4,5
  if (isWithinRegion(pos, REGION_3a)
      || isWithinRegion(pos, REGION_3b)
      || isWithinRegion(pos, REGION_4a)
      || isWithinRegion(pos, REGION_4b)
      || isWithinRegion(pos, REGION_5a)
      || isWithinRegion(pos, REGION_5b))
    return true;
  return false;
}

/*! This method determines which possible shooting point provides
    the greatest probability of scoring and returns that probability.

*/
/*double WorldModel::getBestScoringProbability( VecPosition* scoringPoint)
{

  return 0.75;
} */

/*! This implementation is currently inefficient because it checks all points
    instead of hill climbing.  Variable allocation inside the for loop could
    also use optimizing.

    Scoring log number 600
*/
double WorldModel::getBestScoringProbability( VecPosition* scoringPoint)
{
  Log.logWithTime(600, "Entering getBestScoringProbability\n");
  VecPosition* localBestScoringPoint = new VecPosition(0,0);
  VecPosition ballPos = getBallPos();
  double bestScoringProb = 0;
  double goalEnteringProbability = 0;
  double goaliePassingProb = 0;
  double currentProbability = 0;
  double goalWidth = SS->getGoalWidth();
  VecPosition opponentGoaliePos = getGlobalPosition(OBJECT_OPPONENT_GOALIE);
  VecPosition agentPos = getAgentGlobalPosition();
  VecPosition rightPost = VecPosition(PITCH_LENGTH/2.0, goalWidth/2.0);
  VecPosition leftPost = VecPosition(PITCH_LENGTH/2.0, -goalWidth/2.0);
  double distBetweenBallAndGoalie = ballPos.getDistanceTo(opponentGoaliePos);
  AngDeg  angleBetweenGoalieAndShootingPoint = 0;
  double A = 0;
  double O = 0;
  double alpha = 0;
  double theta = 0;
  double probOutThis = 1;
  double probOutThat = 1;
  VecPosition thisPost;
  VecPosition thatPost;
  Line shootingLine(0,0,0);
  double distanceFromShootingPointToThisPost = 0;
  double distanceFromShootingPointToThatPost = 0;
  double distanceFromBallToShootingPoint = 0;
  VecPosition closestPointOnShootingLine;
  double shortestDistanceFromThisPostToShootingLine = 0;
  double distFromBallToIntersectionOfShortestLine = 0;
  double distFromBallToEdge = (PITCH_LENGTH/2.0)-ballPos.getX();
  //for all discretized possible points in .1 meter intervals (140 loops)
  for (VecPosition* currentPoint = new VecPosition(PITCH_LENGTH/2.0,-goalWidth/2.0); currentPoint->getY() < goalWidth/2.0; currentPoint->setY(currentPoint->getY()+0.1))
  {
  //calculate probability
    //if distance to shooting point > 32m
    if (ballPos.getDistanceTo(*currentPoint) > 32)
    {
      //assign 0 probability of entering goal
      goalEnteringProbability = 0;
//      Log.logWithTime(600, "Further than 32 meters from shooting point, prob = 0\n");
    }
    //else if shooting point and ball position have the same y coordinate.
    else if (fabs(ballPos.getY() - currentPoint->getY()) < EPSILON) //Double precision problems?
    {
      //  use perpendicular method to calculate probability of the ball entering the goal
      goalEnteringProbability = (erf(currentPoint->getDistanceTo(rightPost)/sigma(currentPoint->getDistanceTo(agentPos)))/2.0) + (erf(currentPoint->getDistanceTo(leftPost)/sigma(currentPoint->getDistanceTo(agentPos)))/2.0);
      Log.logWithTime(600, "using perpendicular method. goalEnteringProbability = %f\n", goalEnteringProbability);
      //cout << goalEnteringProbability << endl;
    }
    //else
    else
    {
    //  use other method to calculate probability of ball entering goal
      shootingLine = Line::makeLineFromTwoPoints(ballPos, *currentPoint);
      theta = 0;
      alpha = 0;

      //determine which side is the close post
        //if the y coordinate of the shooting point is greater than the y coord
        //of the ball then the left post is closest.
      if (currentPoint->getY() < ballPos.getY())
      {
        thisPost = leftPost;
        thatPost = rightPost;
      }
        //if the y coordinate of the shooting point is less than the y coord of
        //the ball then the right post is closest.
      else
      {
        thisPost = rightPost;
        thatPost = leftPost;
      }

      distanceFromShootingPointToThisPost = currentPoint->getDistanceTo(thisPost);
      distanceFromShootingPointToThatPost = currentPoint->getDistanceTo(thatPost);
        //calculate distance from ball to shooting point
      distanceFromBallToShootingPoint = currentPoint->getDistanceTo(ballPos);
      //calculate distance from shooting point to this post
      distanceFromShootingPointToThisPost = currentPoint->getDistanceTo(thisPost);



     /* Still working this out....*/

      //calculate shortest distance from this post to shooting line
      closestPointOnShootingLine = shootingLine.getPointOnLineClosestTo(thisPost);
      distFromBallToIntersectionOfShortestLine = closestPointOnShootingLine.getDistanceTo(ballPos);
      if (distanceFromBallToShootingPoint - distFromBallToIntersectionOfShortestLine > EPSILON)
      {
        A = distanceFromBallToShootingPoint - distFromBallToIntersectionOfShortestLine;
        //calculate theta
        theta = asinDeg(A / distFromBallToEdge); //distanceFromBallToShootingPoint is the wrong thing here!
        shortestDistanceFromThisPostToShootingLine = thisPost.getDistanceTo(closestPointOnShootingLine);
        //calculate probability that the ball is out to this side
        probOutThis = (erf(shortestDistanceFromThisPostToShootingLine/sigma(distFromBallToIntersectionOfShortestLine))/2.0);
      }
      else
      {
        //probability that the ball is out to this side = 0
        probOutThis = 0;
        //calculate theta
        Line goalLine = Line::makeLineFromTwoPoints(leftPost, rightPost);
        theta = asinDeg(ballPos.getDistanceTo(goalLine.getPointOnLineClosestTo(ballPos)) / distanceFromBallToShootingPoint);
      }


      /*end work in progress*/

/*
      //If the side adjacent to theta of the theoretical right angle triangle
      //with H and theta is >= the distance from the ball to the shooting point
      //then the ball can't go out to this side, because it can't go backwards.
      if (distanceFromBallToShootingPoint < distanceFromShootingPointToThisPost * cosDeg(theta)) //This isn't right yet either
      {

        probOutThis = 0;
      }
      //else
      else
      {
      }
*/
      //do other side
        //calculate distance from shooting point to this post
        distanceFromShootingPointToThatPost = currentPoint->getDistanceTo(thatPost);
        //calculate A
        A = distanceFromShootingPointToThatPost * cosDeg(theta);
        //calculate O
        O = distanceFromShootingPointToThatPost * sinDeg(theta);
        //calculate the probability that the ball is out to this side
        probOutThat = (erf(O/sigma(A+distanceFromBallToShootingPoint))/2.0);

      goalEnteringProbability = 1 - (probOutThis + probOutThat);
      Log.logWithTime(600, "Using non-perpendicular method, goalEnteringProbability = %f\n",goalEnteringProbability);
      //cout << goalEnteringProbability << endl;
    }
    //calculate probability of the ball passing the goalie
    if (ballPos.getDistanceTo(*currentPoint) <= 32)
    {
      angleBetweenGoalieAndShootingPoint = (AngDeg)fabs((*currentPoint - agentPos).getDirection() - (opponentGoaliePos - agentPos).getDirection());
      goaliePassingProb = 1/(1+exp(-9.5*u(angleBetweenGoalieAndShootingPoint,distBetweenBallAndGoalie)));
      Log.logWithTime(600, "GoaliePassingProb = %f\n", goaliePassingProb);
      //cout << goaliePassingProb << endl;
      //probability = product of passing goalie and entering goal.
      currentProbability = goaliePassingProb * goalEnteringProbability;
      Log.logWithTime(600, "Current Probability = %f\n", currentProbability);
//      cout << currentProbability << endl;
    //do maxima check
      //if the probability is > than the current maximum, make it the current maximum.
      if (currentProbability > bestScoringProb)
      {
        bestScoringProb = currentProbability;
        localBestScoringPoint->setX(currentPoint->getX());
        localBestScoringPoint->setY(currentPoint->getY());
      }
      Log.logWithTime(600, "bestScoringProb = %f\n", bestScoringProb);
      //cout << bestScoringProb << endl;
    }
    else
    {
      goaliePassingProb = 0;
    }
  }
  //set scoring point and return probability
  Log.logWithTime(600, "Exiting getBestScoringProb\n");

  //Stub for getBestScoringProbability
  //cout << "Called stub for getBestScoringProbability!\n";

  if (bestScoringProb > 0) cout << "BestScoringProb = " << bestScoringProb << endl;
  if (localBestScoringPoint != NULL) scoringPoint = localBestScoringPoint;
  return bestScoringProb;
}



/*! This method finds a teammate within the specified region
    it does so in no guaranteed order and if there is more than
    one teammate in the region it does not guarantee which one
    will be returned.

    \param region RegionT describing which region to check in
    \return ObjectT a teammate in the region, OBJECT_ILLEGAL if there is no such teammate

*/
ObjectT WorldModel::getTeammateWithinRegion( RegionT region)
{
  //iterate through teammates until we find one in the specified region.
  int i = 0;
  ObjectT object = OBJECT_ILLEGAL;
  for (object = iterateObjectStart(i, OBJECT_SET_TEAMMATES); object != OBJECT_ILLEGAL; object = iterateObjectNext(i,OBJECT_SET_TEAMMATES))
  {
    if (isObjectWithinRegion(object, region)) return object;
  }
  return OBJECT_ILLEGAL;
}

/*! This method determines if the specified object qualifies
    as very free, based on a freedom threshold.  The object
    must be a PlayerObject for this method to function correctly.

    \return bool true if the object's freedom is higher than the very free threshold

*/
bool WorldModel::isVeryFree(ObjectT object)
{
  if (object != OBJECT_ILLEGAL)
    return ((PlayerObject*)getObjectPtrFromType(object))->getFreedom() > PS->getVeryFreeThr();
  else return false;
}

/*! This method determines if the specified object qualifies
    as fairly free, based on a freedom threshold.  The object
    must be a PlayerObject for this method to function correctly.

    \return bool true if the object's freedom is higher than the fairly free threshold

*/
bool WorldModel::isFairlyFree(ObjectT object)
{
  if (object != OBJECT_ILLEGAL)
    return ((PlayerObject*)getObjectPtrFromType(object))->getFreedom() > PS->getFairlyFreeThr();
  else return false;
}

/*! This method determines if the specified object qualifies
    as free, based on a freedom threshold.  The object
    must be a PlayerObject for this method to function correctly.

    \return bool true if the object's freedom is higher than the free threshold

*/
bool WorldModel::isFree(ObjectT object)
{
  if (object != OBJECT_ILLEGAL)
    return ((PlayerObject*)getObjectPtrFromType(object))->getFreedom() > PS->getFreeThr();
  else return false;
}

double WorldModel::getFreedom(ObjectT object)
{

  if (object != OBJECT_ILLEGAL)
    return ((PlayerObject*)getObjectPtrFromType(object))->getFreedom();
  else return -1000.0;
}

/*! This method supports the scoring problem solution from the
    UvA master's thesis.

*/
double WorldModel::u(double a, double d)
{
  return 0.043*a + 0.09 * d - 2.1323;
}

/*! This method supports the scoring problem solution from the
    UvA master's thesis.

*/
double WorldModel::sigma(double d)
{
  return -1.88* log(1-d/45.0);
}
