/*
Copyright (c) 2000-2002, Jelle Kok, University of Amsterdam
All rights reserved.

Redistribution and use in source and binary forms, with or without 
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
list of conditions and the following disclaimer. 

2. Redistributions in binary form must reproduce the above copyright notice, 
this list of conditions and the following disclaimer in the documentation 
and/or other materials provided with the distribution. 

3. Neither the name of the University of Amsterdam nor the names of its 
contributors may be used to endorse or promote products derived from this 
software without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE 
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
/*! \file WorldModelPredict.C
<pre>
<b>File:</b>          WorldModelPredict.C
<b>Project:</b>       Robocup Soccer Simulation Team: UvA Trilearn
<b>Authors:</b>       Jelle Kok
<b>Created:</b>       12/02/2001
<b>Last Revision:</b> $ID$
<b>Contents:</b>      class definitions of WorldModel. This class contains
               methods that give predictions about future states of the world.
<hr size=2>
<h2><b>Changes</b></h2>
<b>Date</b>             <b>Author</b>          <b>Comment</b>
12/02/2001       Jelle Kok       Initial version created
</pre>
*/

#include <stdio.h>
#include "WorldModel.h"

/*****************************************************************************/
/************************** PREDICTIONS **************************************/
/*****************************************************************************/

/*! This method predicts the state of an player after it performs a specific
    SoccerCommand. The current state of the agent is passed by giving the
    position, velocity and body and neck direction. These arguments are
    updated and after return of this method will contain the new values
    as if the command was performed.
    \param com SoccerCommand that will be performed
    \param pos current position of the object.
    \param vel current velocity of the object
    \param angGlobalBody global body angle
    \param angGlobalBody global neck angle
    \param current stamina information.
    \return boolean which indicates whether values were updated */
bool WorldModel::predictStateAfterCommand( SoccerCommand com,
    VecPosition *pos, VecPosition *vel, AngDeg *angGlobalBody,
    AngDeg *angGlobalNeck, Stamina *sta )
{
  switch( com.commandType ) // based on kind of command, choose action
  {
    case CMD_DASH:
      if( playMode != PM_BEFORE_KICK_OFF )
        predictStateAfterDash( com.dPower, pos, vel, sta, *angGlobalBody );
      break;
    case CMD_TURN:
      predictStateAfterTurn(com.dAngle,pos,vel,angGlobalBody,angGlobalNeck,sta);
      break;
    case CMD_TURNNECK: // note that position and velocity are not updated
      *angGlobalNeck = VecPosition::normalizeAngle(*angGlobalNeck + com.dAngle);
      break;
    case CMD_KICK:
    case CMD_CATCH:
      predictStateAfterDash( 0.0, pos, vel, sta, *angGlobalBody );
      break;
    case CMD_MOVE:
      pos->setVecPosition( com.dX, com.dY );
      vel->setMagnitude( 0.0 );
      break;
    case CMD_ILLEGAL:
      predictStateAfterDash( 0.01, pos, vel, sta, *angGlobalBody );
      break;
    default:
      return false;
      break;
  }
  return true;
}

/*! This method predicts the state of the agent after it performs a specific
    SoccerCommand. This method makes use of the method predictInfoAfterCommand.
    All arguments are initialized with the current values of the agent.
    \param com SoccerCommand that will be performed
    \param pos will be filled with updated position of agent
    \param vel will be filled with updated velocity of agent
    \param angGlobalBody will be filled with global body angle of agent
    \param angGlobalBody will be filled with global neck angle of agent
    \param current will be filled with stamina information of agent
    \return boolean which indicates whether values were updated */
bool WorldModel::predictAgentStateAfterCommand( SoccerCommand com,
    VecPosition *pos, VecPosition *vel, AngDeg *angGlobalBody,
    AngDeg *angGlobalNeck, Stamina *sta )
{
  *pos           = getAgentGlobalPosition();
  *vel           = getAgentGlobalVelocity();
  *angGlobalBody = getAgentGlobalBodyAngle();
  *angGlobalNeck = getAgentGlobalNeckAngle();
  *sta           = getAgentStamina();
  predictStateAfterCommand( com, pos, vel, angGlobalBody, angGlobalNeck, sta );

  return true;
}

/*! This method returns the global position of the agent after the specified
    command is performed. This method makes use of the method
    'predictAgentInfoAfterCommand'
    \param com SoccerCommand that will be performed.
    \return VecPosition new global position of the agent. */
VecPosition WorldModel::predictAgentPosAfterCommand( SoccerCommand com )
{
  VecPosition p1, p2;
  AngDeg      a1, a2;
  Stamina     sta;
  predictAgentStateAfterCommand( com, &p1, &p2, &a1, &a2, &sta );
  return p1;
}

/*! This method determines the state of a player after a dash command is
    performed. The current state of the player is specified by the passed
    arguments. After this method returns, all arguments are updated.
    \param pos initial position, will be changed to the predicted position
    \param vel intital velocity, will be changed to the predicted velocity
    \param dActualPower actual power that is send with dash command
    \param sta pointer to stamina, when not NULL, effort is used and updated
    \param dDirection direction of dash */
void WorldModel::predictStateAfterDash( double dActualPower, VecPosition *pos,
             VecPosition *vel, Stamina *sta, double dDirection )
{
  // get acceleration associated with actualpower
  double dEffort = ( sta != NULL ) ? sta->getEffort() : 1.0;
  double dAcc    = dActualPower * SS->getDashPowerRate() * dEffort;

  // add it to the velocity; negative acceleration in backward direction
  if( dAcc > 0 )
    *vel += VecPosition::getVecPositionFromPolar( dAcc, dDirection );
  else
    *vel += VecPosition::getVecPositionFromPolar( fabs(dAcc),
        VecPosition::normalizeAngle(dDirection+180));

  // check if velocity doesn't exceed maximum speed
  if( vel->getMagnitude() > SS->getPlayerSpeedMax() )
    vel->setMagnitude( SS->getPlayerSpeedMax() );

  // add velocity to current global position and decrease velocity
  *pos += *vel;
  *vel *= SS->getPlayerDecay();
  if( sta != NULL )
    predictStaminaAfterDash( dActualPower, sta );
}

/*! This method determines the state of a player after a turn command is
    performed. The global position is updated with the velocity and the
    velocity is updated. Then the actual turn angle is calculated taken the
    inertia into account. This actual turn angle is used to update both
    the global body and global neck direction.
    \param dSendAngle actual angle given in command
    \param pos initial position, will be changed to the predicted position
    \param vel intital velocity, will be changed to the predicted velocity
    \param angBody global body direction
    \param angNeck global neck direction
    \param sta Stamina of player can be NULL */
void WorldModel::predictStateAfterTurn( AngDeg dSendAngle, VecPosition *pos,
          VecPosition *vel, AngDeg *angBody, AngDeg *angNeck, Stamina *sta )
{
  // calculate effective angle that is turned and turn body and neck accordingly
  double dEffectiveAngle = getActualTurnAngle( dSendAngle, vel->getMagnitude());
  *angBody = VecPosition::normalizeAngle( *angBody + dEffectiveAngle );
  *angNeck = VecPosition::normalizeAngle( *angNeck + dEffectiveAngle );

  // update as if dashed with no power
  predictStateAfterDash( 0.0, pos, vel, sta, *angBody );
  return;
}

/*! This method determines the global position of the object o after iCycles
    When the object is the ball, only the decay of the ball is taken into
    account. When the object is a player it
    is assumed that the player dashes with 'iDashPower' every cycle.
    \param o objectT of which global position will be predicted
    \param iCycles pos is predicted after this number of cycles
    \param iDashPower dash power that is used every cycle in dash (default 100)
    \param vel will be filled with predicted global velocity after iCycles
    \return predicted global position after iCycles. */
VecPosition WorldModel::predictPosAfterNrCycles( ObjectT o, int iCycles,
                                       int iDashPower, VecPosition *velocity )
{
  VecPosition vel = getGlobalVelocity( o );
  VecPosition pos = getGlobalPosition( o );

  if( o == OBJECT_BALL )
  {
    // get the speed and the distance it travels in iCycle's.
    // use this distance and direction it travels in, to calculate new pos
    // geom series is serie s=a+ar+..+ar^n...decay=r,iCycles=n,dSpeed=a
    double dDist = Geometry::getSumGeomSeries( vel.getMagnitude(),
                                               SS->getBallDecay(),
                                               (double)iCycles);
    pos          += VecPosition( dDist, vel.getDirection(), POLAR );
  }
  else if( SoccerTypes::isKnownPlayer( o )  )
  {
    double      dDirection = 0.0; // used when no info about global body
    Stamina     stamina;          // used when object is agent

    if( getAgentObjectType() == o )
    {
      dDirection = getAgentGlobalBodyAngle();
      stamina    = getAgentStamina();
    }
    else if( getTimeGlobalAngles(o) > getCurrentTime() - 2 )
      dDirection = getGlobalBodyAngle(o);

    for( int i = 0; i < iCycles ; i ++ )
      predictStateAfterDash( iDashPower, &pos, &vel, &stamina, dDirection );
  }
  if( velocity != NULL )
    *velocity = vel;
  return pos;
}

/*! This method predicts the position of the agent after 'iCycles' when every
    cycle is dashed with 'iDashPower'. The method
    'predictGlobalPosAfterNrCycles' is used to calculate this position.
    \param iCycles number of cycles
    \param iDashPower dash power that is passed
    \return VecPosition indicating predicted global position of agent. */
VecPosition WorldModel::predictAgentPos( int iCycles, int iDashPower )
{
  return predictPosAfterNrCycles( getAgentObjectType(), iCycles, iDashPower);
}

/*! This method gives an estimate for the number of cycles it will a player to
    reach a specific position. A position is reached when the player is located
    in the maximal kick distance of this position. When this is not the case
    a dash (or turn) is performed until the player is in the kickable distance.
    \param o objectT which wants to reach posTo
    \param posTo global position which has to be reached
    \param angToTurn angle to 'posTo' when is turned (instead of dashed)
    \return predicted nr of cycles for o to reach posTo  */
int WorldModel::predictNrCyclesToPoint( ObjectT o, VecPosition posTo,
                                        AngDeg angToTurn )
{
  VecPosition posGlobal        = getGlobalPosition( o );
  AngDeg      angBody;
  AngDeg      ang;

  if( getAgentObjectType() == o )
    angBody = getAgentGlobalBodyAngle();
  else if( fabs(double(getTimeGlobalAngles(o)-getCurrentTime())) < 2 )
    angBody = getGlobalBodyAngle( o );
  else
    angBody = 180;

  int iExtraCycles;
  ang = fabs(angBody-(posTo-posGlobal).getDirection());
  if( ang > 20 && ang < 100 )
    iExtraCycles = 1;
  else if( ang > 100 )
    iExtraCycles = 2;
  else
    iExtraCycles = 0;

  if( posGlobal.getDistanceTo( posTo ) < SS->getMaximalKickDist() )
    return 0;
  else
    return iExtraCycles + int((posGlobal.getDistanceTo( posTo )
           /min(1.0, SS->getPlayerSpeedMax())));
}

/*! This method returns the number of cycles it will take the object 'objFrom'
    to reach the object 'objTo' (usually respectively the player and the ball).
    \param objFrom ObjectT that is the object that wants to move
    \param objTo ObjectT to which is moved
    \return number of cycles it will take objFrom to move to objTo */
int WorldModel::predictNrCyclesToObject( ObjectT objFrom, ObjectT objTo )
{
  VecPosition posPrev(UnknownDoubleValue,UnknownDoubleValue);
  
  if( objFrom == OBJECT_ILLEGAL || objTo == OBJECT_ILLEGAL ||
      getGlobalPosition( objFrom ).getDistanceTo( getGlobalPosition( objTo )
      ) > 40 )
    return 101;

  int         iCycles      = 0;
  int         iCyclesToObj = 100;
  VecPosition posObj(0,0);
 
  // continue calculating number of cycles to position until or we can get
  // earlier at object position, are past maximum allowed number of cycles or
  // the object does not move anymore.
  while( iCycles <= iCyclesToObj && iCycles < PS->getPlayerWhenToIntercept() &&  
         posObj.getDistanceTo( posPrev ) < EPSILON )
  {
    iCycles      = iCycles + 1  ;
    posPrev      = posObj;    
    posObj       = predictPosAfterNrCycles( objTo,   iCycles );
    iCyclesToObj = predictNrCyclesToPoint ( objFrom, posObj,
                                           PS->getPlayerWhenToTurnAngle() );
  }
  if( iCycles == PS->getPlayerWhenToIntercept() )
    printf( "(WorldModel::predictNrCyclesToObject) Too much cycles??\n" );

  return iCyclesToObj;
}



/*! This method updates all the stamina variables using the calculations from
    the soccer manual. It is not really important since stamina is read from
    sense_body every cycle. That information is more up to date.
    \param power of last dash command
    \param stamina pointer to all stamina values, will change to new value
    \return stamina class will be updated to new stamina values */
void WorldModel::predictStaminaAfterDash( double dPower, Stamina *stamina )
{
  double sta = stamina->getStamina();
  double eff = stamina->getEffort();
  double rec = stamina->getRecovery();

  // double negative value when dashed backwards
  sta -= ( dPower > 0.0 ) ? dPower : -2*dPower ;
  if( sta < 0 ) sta = 0;

  // stamina below recovery threshold, lower recovery
  if( sta <= SS->getRecoverDecThr()*SS->getStaminaMax() &&
                         rec > SS->getRecoverMin() )
      rec -= SS->getRecoverDec();

  // stamina below effort decrease threshold, lower effort
  if( sta <= SS->getEffortDecThr()*SS->getStaminaMax() &&
                   eff > SS->getEffortMin() )
      eff -= SS->getEffortDec();

  // stamina higher than effort incr threshold, raise effort and check maximum
  if( sta >= SS->getEffortIncThr() * SS->getStaminaMax() &&
      eff < 1.0)
  {
    eff += SS->getEffortInc();
    if ( eff > 1.0 )
       eff = 1.0;
  }

  // increase stamina with (new) recovery value and check for maximum
  sta += rec*SS->getStaminaIncMax();
  if ( sta > SS->getStaminaMax() )
    sta = SS->getStaminaMax();

  stamina->setStamina ( sta );
  stamina->setEffort  ( eff );
  stamina->setRecovery( rec );
}

/*! This method determines whether a dash command (supplied as the first
    argument) will result in collision with another player.
    This is checked by determing the global position after the command
    and then check whether the positions of one of the other players lies
    with the player size. Since it cannot be known what kind of action the
    other player takes in this cycle, it is also difficult to predict what the
    global position of the player will be in the next cycle. This method
    therefore assumes the other players have issued a dash with maximum power
    in the last cycle.
    \return bool indicating whether dash will result in a collision. */
bool WorldModel::isCollisionAfterDash( SoccerCommand soc )
{
  if( soc.commandType != CMD_DASH )
    return false;
                                             
  VecPosition posPred = predictAgentPosAfterCommand( soc );
  VecPosition posBall = predictPosAfterNrCycles( OBJECT_BALL, 1 );
  if( posPred.getDistanceTo( posBall ) < 
              SS->getPlayerSize() + SS->getBallSize()  )
   return true;
   
  return false;
}


/*! This method returns a Vector teammates that have a freedom score greater
    than a minimum threshold.  The Vector is in descending order from most free
    to least.

    \return vector< ObjectT > of free teammates in order from most to least free
    \author David Brokenshire
    ***possibly buggy or defective***
*/
vector< ObjectT > WorldModel::getFreeTeammateVector()
{
  //for all known teammates
  int i = 0;
  vector<ObjectT> sortedByFreedom;
  double lastFreedomScore = -1000;
  double freedomScore = -1000;
  vector<ObjectT>::iterator lastAddition;
  for (ObjectT object = iterateObjectStart(i, OBJECT_SET_TEAMMATES); object != OBJECT_ILLEGAL; object = iterateObjectNext(i, OBJECT_SET_TEAMMATES))
  {
    // exclude goalie
    if (object == OBJECT_TEAMMATE_1)
      continue;
    // exclude players with low confidence
   // if (getConfidence( object) < PS->getPlayerConfThr())
   if (!isConfidenceVeryGood(object))
      continue;
    lastFreedomScore = freedomScore;
    //assign freedom score
    freedomScore = calculateFreedom(object);
    if (sortedByFreedom.size() == 0)
    {
      sortedByFreedom.push_back(object);
      lastAddition = sortedByFreedom.begin();
    }
    else if (freedomScore >= lastFreedomScore)
      lastAddition = sortedByFreedom.insert(lastAddition, object);
    else
      lastAddition = sortedByFreedom.insert(lastAddition+1, object); //I'm not sure if I'm using iterators right here.
  }

  //return sorted vector
  return sortedByFreedom;
}

/*! This method returns a Vector teammates that have a freedom score greater
    than a minimum threshold.  The Vector is in descending order from most free
    to least.

    \return vector< ObjectT > of free teammates in order from most to least free
    \author David Brokenshire
    ***possibly buggy or defective***
*/
vector< ObjectT > WorldModel::getFreeTeammateVectorP(VecPosition point)
{
  //for all known teammates
  int i = 0;
  vector<ObjectT> sortedByFreedom;
  double lastFreedomScore = -1000;
  double freedomScore = -1000;
  vector<ObjectT>::iterator lastAddition;
  for (ObjectT object = iterateObjectStart(i, OBJECT_SET_TEAMMATES); object != OBJECT_ILLEGAL; object = iterateObjectNext(i, OBJECT_SET_TEAMMATES))
  {
    lastFreedomScore = freedomScore;
    //assign freedom score
    freedomScore = calculateFreedomFromPoint(object, point);
    if (sortedByFreedom.size() == 0)
    {
      sortedByFreedom.push_back(object);
      lastAddition = sortedByFreedom.begin();
    }
    else if (freedomScore >= lastFreedomScore)
      lastAddition = sortedByFreedom.insert(lastAddition, object);
    else
      lastAddition = sortedByFreedom.insert(lastAddition+1, object); //I'm not sure if I'm using iterators right here.
  }

  //return sorted vector
  return sortedByFreedom;
}

/*! This method calculates the freedom value for the specified object
    Freedom is calculated as some weird hacked together formula that
    probably doesn't mean much and should be fixed.



double WorldModel::calculateFreedom(ObjectT object)
{
  double freedom = 0;
  if (object == OBJECT_ILLEGAL)
    return -1000;
  PlayerObject* player = (PlayerObject*)getObjectPtrFromType(object);
  VecPosition ballPos = getBallPos();
  VecPosition playerPos = player->getGlobalPosition();

  //Basic version - Three possible cone widths
  if (getNrInSetInCone(OBJECT_SET_OPPONENTS,1,ballPos,playerPos) == 0)
    freedom = 1;
  else if (getNrInSetInCone(OBJECT_SET_OPPONENTS,0.7,ballPos,playerPos) == 0)
    freedom = 0.7;
  else if (getNrInSetInCone(OBJECT_SET_OPPONENTS,0.45,ballPos,playerPos) == 0)
    freedom = 0.5;


  player->setFreedom(freedom);
  return freedom;
} */

/*! A slightly improved version of calculating the freedom value. Takes into account the
    distance to the closest opponent and distance to the ball.
    \author S.Razykov
    ***magic numbers***
*/
double WorldModel::calculateFreedom(ObjectT object)
{
  double freedom = 0.0;
  if (object == OBJECT_ILLEGAL)
    return -1000;
  PlayerObject* player = (PlayerObject*)getObjectPtrFromType(object);
  VecPosition ballPos = getBallPos();
  //VecPosition playerPos = player->getGlobalPosition();
  VecPosition playerPos = getBestPointForDirectPass(object);
  VecPosition opponentPos = predictPlayerPos(3, getClosestInSetTo(OBJECT_SET_OPPONENTS, ballPos));

  //Basic version - Three possible cone widths
  if (getNrInSetInCone(OBJECT_SET_OPPONENTS,1.0,ballPos,playerPos) == 0)
    freedom = 1.0;
  else if (getNrInSetInCone(OBJECT_SET_OPPONENTS,0.9,ballPos,playerPos) == 0)
    freedom = 0.9;
  else if (getNrInSetInCone(OBJECT_SET_OPPONENTS,0.8,ballPos,playerPos) == 0)
    freedom = 0.8;
  if (getNrInSetInCone(OBJECT_SET_OPPONENTS,0.7,ballPos,playerPos) == 0)
    freedom = 0.7;
  else if (getNrInSetInCone(OBJECT_SET_OPPONENTS,0.6,ballPos,playerPos) == 0)
    freedom = 0.6;
  else if (getNrInSetInCone(OBJECT_SET_OPPONENTS,0.3,ballPos,playerPos) == 0)
    freedom = 0.5;  

  if (freedom  > 0.0)
  {
  // distance to the opponnent
  if( playerPos.getDistanceTo(opponentPos) > 6.0)
    freedom = freedom+1.0;
  else if( playerPos.getDistanceTo(opponentPos) > 5.0)
    freedom = freedom+0.9;  
  else if(  playerPos.getDistanceTo(opponentPos) > PS->getOppCloseThreshold())
    freedom  = freedom + 0.8;
  else if(  playerPos.getDistanceTo(opponentPos) > PS->getOppVeryCloseThreshold())
    freedom  =freedom + 0.5;
  else
    freedom  = freedom + 0.0;
  }

  if (freedom > 0.0)
  {
  // distance to the ball
  if( playerPos.getDistanceTo(ballPos) <30.0 )
    freedom = freedom+1.0;
  else if( playerPos.getDistanceTo(ballPos) < 36.0 )
    freedom  = freedom + 0.8;
  else if( playerPos.getDistanceTo(ballPos) < 38.0 )
   freedom  =freedom + 0.6;
  else
    freedom  = freedom + 0.0;
  }

  // reduce freedom for back passes
  if( playerPos.getX() < ballPos.getX() - 18.0 )
    freedom = freedom - 1.0;
  else if( playerPos.getX() < ballPos.getX() - 10.0 )
    freedom  = freedom - 0.5;
  else if(  playerPos.getX() < ballPos.getX() - 6.0 )
   freedom  =freedom - 0.4;
  else
    freedom  = freedom - 0.0;

  // take average
  freedom = freedom/3.0;

  player->setFreedom(freedom);
  //cout<<"freedom "<<freedom<<endl;
  return freedom;
}  


/*! This method calculates is the same as privious, but take to respect any point
    in the field
*/
double WorldModel::calculateFreedomFromPoint(ObjectT object, VecPosition point)
{
  double freedom = 0;
  if (object == OBJECT_ILLEGAL)
    return -1000;
  PlayerObject* player = (PlayerObject*)getObjectPtrFromType(object);
  VecPosition ballPos = getBallPos();
  VecPosition playerPos = point;

  //Basic version - Three possible cone widths
  if (getNrInSetInCone(OBJECT_SET_OPPONENTS,1,ballPos,playerPos) == 0)
    freedom = 1;
  else if (getNrInSetInCone(OBJECT_SET_OPPONENTS,0.7,ballPos,playerPos) == 0)
    freedom = 0.7;
  else if (getNrInSetInCone(OBJECT_SET_OPPONENTS,0.45,ballPos,playerPos) == 0)
    freedom = 0.5;


  player->setFreedom(freedom);
  return freedom;
}


/*! This method returns the center of the target area, wich is a part of the WorldModel*/

VecPosition  WorldModel::getTargetArea( )
{
  return targetArea;
}

void  WorldModel::setTargetArea( VecPosition targetAreaCenter )
{
  targetArea = targetAreaCenter;  // set center
  
  // define rectangle
  t_Rectangle->setRectanglePoints(VecPosition(targetArea.getX()-PITCH_LENGTH/X_CELLS*NX_CELLS/2,
                                  targetArea.getY()-PITCH_WIDTH/Y_CELLS*NY_CELLS/2),
                                  VecPosition( targetArea.getX() + PITCH_LENGTH/X_CELLS*NX_CELLS/2 ,
                                  targetArea.getY() + PITCH_WIDTH/Y_CELLS*NY_CELLS/2));
  // set rectangle of the target area
  t_Region.setArea( t_Rectangle);
  
}

bool WorldModel::isTargetAreaOffence()
{ return targetAreaOffence;
}

/*! this metod sets parameters for target area - the center, rectangle and region
    when the team  posesses the ball
*/
VecPosition  WorldModel::setTargetAreaOffense( )

{

  
  // initial and current values
  int iCycles           =  P_TIME;
  int curri             =     0;
  int currj             =     0;
  double curraverage    =     0.0;
  double myaverage      =   UnknownDoubleValue;

  int attackInt         =  getXIndexCellWithBall( );  // current X index of the cell with the ball
  int yIndex            =  getYIndexCellWithBall( );  // current X index of the cell with the ball
  bool insideT          = false;                      // if the ball is inside "the last" target area

  // indexes
  int i;
  int j;
  int m;
  int n;

  computeReachToCells( iCycles, true );

  // here the hidden transformation from double to int is used, may be it is not very good

  if (attackInt < (X_CELLS-NX_CELLS/2)-NX_CELLS)   //before the last target area
    attackInt = attackInt+NX_CELLS;          //check next area
  else if (attackInt <  X_CELLS-NX_CELLS )  //between the last target area and center of the prvious TA
    attackInt = attackInt+NX_CELLS/2+1;  // the last target area only
  else                                  //inside last target area
  {
    //attackInt = attackInt-NX_CELLS/2-1; // the last and previous target area
    attackInt = attackInt-NX_CELLS/2; // the last target area
    insideT = true;
  }


  for ( i = attackInt ; i < min((attackInt+NX_CELLS),X_CELLS-NX_CELLS/2); i++ )  //starts count cells
  {
    for ( j = NY_CELLS/2; j < (Y_CELLS-NY_CELLS/2); j++ )
    {
      
      if ( abs(yIndex-j)>NY_CELLS) //cells too far by wide
      {
       continue;
      }

      if (insideT && abs(yIndex-j)<NY_CELLS) //if last target exclude the same Target Area
      {
        continue;
      }
 
    // inner loop for all cells in rectangle with size of target area and center in a[i][j]
      for ( m = ( i - NX_CELLS/2 ); m <=  ( i + NX_CELLS/2 ); m++ )  // starts count cells
      {
        for ( n = ( j -  NY_CELLS/2 ); n <=  ( j + NY_CELLS/2 ); n++ )
        {
          //curraverage = curraverage + ( (Cells[m][n].getProbInCell().getX()+1.0)/(Cells[m][n].getProbInCell().getY()+1.0) );
          curraverage = curraverage + ( (Cells[m][n].getProbInCell().getX()+1.0)-(Cells[m][n].getProbInCell().getY()+1.0) );
        }
      }
      if ( curraverage > myaverage )
      {
        myaverage = curraverage;
        curri = i;
        currj = j;
      }
      // choose area closer to the opponent's goal
      else if (myaverage - curraverage < 2)  // magic number
      {
         if (getBallPos().isInFrontOf(-PITCH_LENGTH/4))
         {
          if (Cells[curri][currj].getCentOfCell().getDistanceTo( getPosOwnGoal()) <
             Cells[i][j].getCentOfCell().getDistanceTo( getPosOwnGoal()) )
           {
             myaverage = curraverage;
             curri = i;
             currj = j;
           }
         }
         else
         {
          if (Cells[curri][currj].getCentOfCell().getDistanceTo( getPosOpponentGoal()) >
             Cells[i][j].getCentOfCell().getDistanceTo( getPosOpponentGoal()) )
           {
             myaverage = curraverage;
             curri = i;
             currj = j;
           }
         }  
      }

    }


  }

  return  Cells[curri][currj].getCentOfCell();
  //targetArea =  Cells[curri][currj].getCentOfCell();
  
  // length and width of the target area
  //double tarLength = PITCH_LENGTH/X_CELLS*NX_CELLS/2;
  //double tarWidth  = PITCH_WIDTH/Y_CELLS*NY_CELLS/2;

  //t_Rectangle->setRectanglePoints(VecPosition(targetArea.getX()-tarLength ,targetArea.getY()-tarWidth),
   //                  VecPosition( targetArea.getX() + tarLength ,targetArea.getY() + tarWidth));
  // set rectangle of the target area
  //t_Region.setArea( t_Rectangle);

  //debug information

  //cout<<"BP "<<getBallPos()<<"ind "<<getXIndexCellWithBall()<<"att "<<attackInt;
  //cout<<" Target = "<<targetArea<<" ";
  //cout<<Cells[curri][currj].getProbInCell()<<endl;

  //return ;

}


/*! this metod sets parameters for target area - the center, rectangle and region
    when the team does not posesses the ball
*/
VecPosition  WorldModel::setTargetAreaDefense( )

{

  
  // initial and current values
  int iCycles           =  P_TIME;
  int curri             =     0;
  int currj             =     0;
  double curraverage    =     0.0;
  double myaverage      =   1000.0;

  int attackInt         =  getXIndexCellWithBall( );  // current X index of the cell with the ball
  int yIndex            =  getYIndexCellWithBall( );  // current X index of the cell with the ball
  bool insideT          = false;                      // if the ball is inside "the last" target area

  // indexes
  int i;
  int j;
  int m;
  int n;

  computeReachToCells( iCycles, false );

  // here the hidden transformation from double to int is used, may be it is not very good

  if (attackInt >= NX_CELLS)   //before the last target area
    attackInt = attackInt-NX_CELLS+2;          //check next area
  else   // too close to own goal
  {
    attackInt = attackInt+NX_CELLS/2+1;  // the last target area only
 // else                                  //inside last target area
 // {
    //attackInt = attackInt-NX_CELLS/2-1; // the last and previous target area
 //   attackInt = attackInt-NX_CELLS/2; // the last target area
    insideT = true;
  }
 // }


  for ( i = max(NX_CELLS/2, attackInt-NX_CELLS) ; i < attackInt; i++ )  //starts count cells
  {                                   
    for ( j = NY_CELLS/2; j < (Y_CELLS-NY_CELLS/2); j++ )
    {
      
      if ( abs(yIndex-j)>NY_CELLS) //cells too far by wide
      {
       continue;
      }

      if (insideT && abs(yIndex-j)<NY_CELLS) //if last target exclude the same Target Area
      {
        continue;
      }

     // inner loop for all cells in rectangle with size of target area and center in a[i][j]
      for ( m = ( i - NX_CELLS/2 ); m <=  ( i + NX_CELLS/2 ); m++ )  // starts count cells
      {
        for ( n = ( j -  NY_CELLS/2 ); n <=  ( j + NY_CELLS/2 ); n++ )
        {
          curraverage = curraverage + ( (Cells[m][n].getProbInCell().getX()+1.0)-(Cells[m][n].getProbInCell().getY()+1.0) );
        }
      }
      if ( curraverage < myaverage )
      {
        myaverage = curraverage;
        curri = i;
        currj = j;
      }
      else if ( curraverage - myaverage< 1)  // magic number
      {
         if (Cells[curri][currj].getCentOfCell().getDistanceTo( getPosOwnGoal()) >
            Cells[i][j].getCentOfCell().getDistanceTo( getPosOwnGoal()) )
         {
           myaverage = curraverage;
           curri = i;
           currj = j;
         }

      }

    }


  }

  return  Cells[curri][currj].getCentOfCell();

  // length and width of the target area
  //double tarLength = PITCH_LENGTH/X_CELLS*NX_CELLS/2;
  //double tarWidth  = PITCH_WIDTH/Y_CELLS*NY_CELLS/2;

  //t_Rectangle->setRectanglePoints(VecPosition(targetArea.getX()-tarLength ,targetArea.getY()-tarWidth),
  //                   VecPosition( targetArea.getX() + tarLength ,targetArea.getY() + tarWidth));
  // set rectangle of the target area
  //t_Region.setArea( t_Rectangle);

  //debug information

  //cout<<"BP "<<getBallPos()<<"ind "<<getXIndexCellWithBall()<<"att "<<attackInt;
  //cout<<" Target = "<<targetArea<<" ";
  //cout<<Cells[curri][currj].getProbInCell()<<endl;

  //return targetAreaCenter;

}

void  WorldModel::computeReachToCells( int iCycles, bool offense  )

{

  // indexes

  int         i     ;
  int         j     ;
  int         m     ;

  double them = 0.0 ;
  double us   = 0.0 ;
  VecPosition p     ;
  int attackInt         =  getXIndexCellWithBall( );
  int startX;
  int finishX;
  
  // sets range of cells for count for offense or defense
  if (offense)    // offense
  {
    if (attackInt > NX_CELLS)  
    {
      startX = attackInt-NX_CELLS;
    }
    else
    {
      startX = attackInt;
    }
    finishX = X_CELLS;
  }
  else   // defense
  {
    startX = 0;
    
    if (attackInt < X_CELLS-NX_CELLS)
    {
      finishX = attackInt+NX_CELLS;
    }
    else
    {
      finishX = attackInt;
    }
  }
    
  for ( j = 0 ; j < Y_CELLS; j++ ) 
  {
    for ( i = startX; i < finishX; i++ )
    {
      for (m  =  1; m < MAX_TEAMMATES ; m ++ )   // for all teammmates exept goalie
      {
        us   = us    + reachFactor( predictNrCyclesToCell( Teammates[m].getType(),
                       Cells[i][j]), iCycles);               
      }
      for (m  =  1; m < MAX_OPPONENTS ; m ++ )   // for all opponents exept goalie
      {
                              
       them = them  + reachFactor( predictNrCyclesToCell( Opponents[m].getType(),
                       Cells[i][j]), iCycles);
                
      }                 
      p.setVecPosition( us, them);
      Cells[i][j].setProbInCell( p );
      
      us  =   0.0;
      them  = 0.0;
    } //end for by j
  }  // end for by i

} // computeReachToCells


double WorldModel::reachFactor( int nrCycles, int iCycles  )

{

  //double parameter = 0.5;

  double parameter = 1.0;

  double nCycl  = static_cast<double>(nrCycles);
  double ratio  = nCycl/iCycles;
  double reach ;
    
  if (  ratio >= 1.0 )        // the player will not reach the cell
  {
    reach = 0.0;
  }
  else if ( ratio < 0.001  )  // if the player is inside
  {
    reach = 2.0;
  }
  else if ( ratio < parameter  )    
  {
    reach = 1.0;
  }
  else
  {
    reach = 0.0;
    //reach = 1.0  - (ratio  - parameter)/(1.0  - parameter);
    //debug inform
    //cout<<"reach factor "<<reach<<" ratio "<<ratio<<endl;
  }
  return reach;
}


int WorldModel::predictNrCyclesToCell( ObjectT o, CellObject cellTo)
{
  VecPosition posGlobal        = getGlobalPosition( o );
  AngDeg      angBody;
  AngDeg      ang;
  double      minDist = cellTo.minDistToCell(posGlobal);

            
  if( getAgentObjectType() == o )
  {
    angBody = getAgentGlobalBodyAngle();
    posGlobal = getAgentGlobalPosition();
  }
  else if( fabs(double(getTimeGlobalAngles(o)-getCurrentTime())) < 2 )
    angBody = getGlobalBodyAngle( o );
  else
    angBody = 180;

  int iExtraCycles;
  ang = fabs(angBody-(cellTo.getCentOfCell()-posGlobal).getDirection());
  if( ang > 20 && ang < 100 )
    iExtraCycles = 1;
  else if( ang > 100 )
    iExtraCycles = 2;
  else
    iExtraCycles = 0;

  if( posGlobal.getDistanceTo( cellTo.getCentOfCell() ) < cellTo.getSizeOfCell().getX() )
    return 0;
  else
    return iExtraCycles + int(((posGlobal.getDistanceTo( cellTo.getCentOfCell() )-minDist)
           /min(1.0, SS->getPlayerSpeedMax())));
}

/*! This method calculates X index of the cell with the ball
*/
int WorldModel::getXIndexCellWithBall( )
{
  int i=0;
  int j;
  double dist = 1000.0;
  double dist1;
  double bX=getBallPos().getX() ;

  for (j=0; j<X_CELLS; j++)
  {
    dist1= fabs(bX-Cells[j][0].getCentOfCell().getX());
    if (dist1 < dist)
    {
      dist = dist1;
      i=j;
    } // endif
  } // endfor 
      
  //cout<<"X ind = "<<i<<endl;
   
  return i;
}

/*! This method calculates Y index of the cell with the ball
*/
int WorldModel::getYIndexCellWithBall( )
{
  int i=0;
  int j;
  double dist = 1000.0;
  double dist1;
  double bY=getBallPos().getY() ;

  for (j=0; j<X_CELLS; j++)
  {
    dist1= fabs(bY-Cells[0][j].getCentOfCell().getY());
    if (dist1 < dist)
    {
      dist = dist1;
      i=j;
    } // endif
  } // endfor

  //cout<<"Y ind = "<<i<<endl;

  return i;
}



void  WorldModel::printDebug()
{
  int i;
  for (i=0; i<MAX_TEAMMATES ; i ++)
  {
    cout<<"# "<<i<<" "<<Opponents[i].getGlobalPosition();
    cout<<" conf "<<getConfidence(Opponents[i].getType());
  }
  cout<<endl;
  cout<<endl;
}

/*! this method returns the length of the target area
*/
double  WorldModel::getLengthOfTarget()
{
  return Cells[0][0].getSizeOfCell().getX()*NX_CELLS;
}

/*! This method returns the heigth of the target area
*/
double  WorldModel::getHeigthOfTarget()
{
  return Cells[0][0].getSizeOfCell().getY()*NY_CELLS;
}


/*! This method returns the average of teammate positions
    \return average position of all the teammates
*/
VecPosition WorldModel::getCenterOfFormation( )
{
  

  // take average of locations
  double x = 0, y = 0;
  for( int i = 0; i < MAX_TEAMMATES  ; i ++ )
  {
     x += Teammates[ i ].getGlobalPosition().getX( );
     y += Teammates[ i ].getGlobalPosition().getY( );
  }

  return VecPosition( x/MAX_TEAMMATES, y/MAX_TEAMMATES );
}

// this method predicts a position of a player assuming that the player will
// keep the same speed

VecPosition WorldModel::predictPlayerPos(int iCycles, ObjectT o)
{
  VecPosition vel = getGlobalVelocity( o );
  VecPosition pos = getGlobalPosition( o );

  // get the speed and the distance it travels in iCycle's.
  // use this distance and direction it travels in, to calculate new pos
    
  double dDist = vel.getMagnitude()*(double)iCycles;
  pos          += VecPosition( dDist, vel.getDirection(), POLAR );
  return pos;
}
